from typing import Any, Dict, Optional, Union

from dlt.sources.helpers import requests

BASE_URL = "https://cloud.getdbt.com/api"


class InvalidCredentialsException(Exception):
    pass


class DBTCloudClientV2:
    """
    A client for interacting with the dbt Cloud API (version 2).
    """

    def __init__(
        self,
        api_token: str,
        account_id: Optional[str] = None,
        base_api_url: str = BASE_URL,
    ):
        """
        Args:
            api_token (str): The API token for authentication.
            account_id (str, optional): The ID of the dbt Cloud account.
                Defaults to None.
            base_api_url (str, optional): The base URL of the dbt Cloud API.
                Defaults to "https://cloud.getdbt.com/api".
        """
        self.api_version = "v2"
        self.base_api_url = f"{base_api_url}/{self.api_version}"
        self._api_token = api_token
        self._headers = {
            "Content-Type": "application/json",
            "Authorization": f"Token {self._api_token}",
            "Accept": "application/json",
        }

        self.account_id = account_id
        self.accounts_url = f"accounts/{self.account_id}"

    def get_endpoint(self, endpoint: str) -> Any:
        response = requests.get(f"{self.base_api_url}/{endpoint}", headers=self._headers)
        results = response.json()
        return results

    def post_endpoint(self, endpoint: str, json_body: Optional[Dict[Any, Any]] = None) -> Any:
        response = requests.post(
            f"{self.base_api_url}/{endpoint}",
            headers=self._headers,
            json=json_body,
        )
        results = response.json()
        return results

    def trigger_job_run(
        self, job_id: Union[int, str], data: Optional[Dict[Any, Any]] = None
    ) -> int:
        """
         Trigger a job run in dbt Cloud.

        Uses `https://cloud.getdbt.com/api/v2/accounts/{account_id}/jobs/{job_id}/run/`
        endpoint to kick off a run for a job.

        When this endpoint returns a successful response, a new run will be enqueued
        for the account. Users can poll the Get run endpoint to poll the run until it completes.
        After the run has completed, users can use the Get run artifact endpoint to download artifacts generated by the run.

        More info: https://docs.getdbt.com/dbt-cloud/api-v2#/operations/Trigger%20Job%20Run.

         Args:
             job_id (int | str): The ID of the job.

             data (dict, optional): The JSON data to post. Defaults to None.
                 Fields of data:
                 '{
                    "cause": "string",
                    "git_sha": "string",
                    "git_branch": "string",
                    "azure_pull_request_id": integer,
                    "github_pull_request_id": integer,
                    "gitlab_merge_request_id": integer,
                    "schema_override": "string",
                    "dbt_version_override": "string",
                    "threads_override": integer,
                    "target_name_override": "string",
                    "generate_docs_override": boolean,
                    "timeout_seconds_override": integer,
                    "steps_override": [
                        "string"
                    ]
                }'

         Returns:
             int: The ID of the triggered job run.

         Raises:
             InvalidCredentialsException: If account_id or job_id is missing.

        """
        if not (self.account_id and job_id):
            raise InvalidCredentialsException(
                "`account_id` and `job_id` are required. "
                f"Got `account_id={self.account_id}` and `{job_id=:}`"
            )

        json_body = {}
        if data:
            json_body.update(data)

        response = self.post_endpoint(f"{self.accounts_url}/jobs/{job_id}/run", json_body=json_body)
        return int(response["data"]["id"])

    def get_run_status(self, run_id: Union[int, str]) -> Dict[Any, Any]:
        """
        Get the status of a dbt Cloud job run by run_id.

        Uses `https://cloud.getdbt.com/api/v2/accounts/{account_id}/runs/{id}/`
        to get job run information.

        More info: https://docs.getdbt.com/dbt-cloud/api-v2#/operations/Retrieve%20Run.

        Args:
            run_id (int | str): The ID of the job run.

        Returns:
            Dict[Any, Any]: The status information of the job run.

        Raises:
            InvalidCredentialsException: If account_id or run_id is missing.

        """
        if not (self.account_id and run_id):
            raise InvalidCredentialsException(
                "`account_id` and `run_id` are required. "
                f"Got `account_id={self.account_id}` and `{run_id=:}`"
            )

        response = self.get_endpoint(f"{self.accounts_url}/runs/{run_id}")
        return dict(response["data"])



"""Template for building a `dlt` pipeline to ingest data from a REST API."""

import dlt
from dlt.sources.rest_api import rest_api_resources
from dlt.sources.rest_api.typing import RESTAPIConfig


# if no argument is provided, `access_token` is read from `.dlt/secrets.toml`
@dlt.source
def rest_api_source(access_token: str = dlt.secrets.value):
    """Define dlt resources from REST API endpoints."""
    config: RESTAPIConfig = {
        "client": {
            # TODO set base URL for the REST API
            "base_url": "https://example.com/v1/",
            # TODO configure the right authentication method or remove
            "auth": {"type": "bearer", "token": access_token},
        },
        "resources": [
            # TODO define resources per endpoint
        ],
        # set `resource_defaults` to apply configuration to all endpoints
    }

    yield from rest_api_resources(config)


pipeline = dlt.pipeline(
    pipeline_name="rest_api_ingest",
    destination="duckdb",
    # `refresh="drop_sources"` ensures the data and the state is cleaned
    # on each `pipeline.run()`; remove the argument once you have a
    # working pipeline.
    refresh="drop_sources",
    # show basic progress of resources extracted, normalized files and load-jobs on stdout
    progress="log",
)


if __name__ == "__main__":
    load_info = pipeline.run(rest_api_source())
    print(load_info)  # noqa: T201



from abc import ABC, abstractmethod
from typing import (
    Any,
    Callable,
    ClassVar,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Set,
    Protocol,
    Type,
)
from dlt.common.libs.sqlglot import TSqlGlotDialect
from dlt.common.destination.configuration import ParquetFormatConfiguration
from dlt.common.exceptions import TerminalValueError
from dlt.common.normalizers.typing import TNamingConventionReferenceArg
from dlt.common.typing import TLoaderFileFormat, get_args
from dlt.common.configuration.utils import serialize_value
from dlt.common.configuration import configspec
from dlt.common.configuration.specs import ContainerInjectableContext
from dlt.common.destination.exceptions import (
    DestinationIncompatibleLoaderFileFormatException,
    DestinationLoadingViaStagingNotSupported,
    DestinationLoadingWithoutStagingNotSupported,
)
from dlt.common.destination.typing import PreparedTableSchema
from dlt.common.arithmetics import DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE
from dlt.common.schema.typing import (
    TColumnSchema,
    TColumnType,
    TTableSchema,
    TLoaderMergeStrategy,
    TTableFormat,
    TLoaderReplaceStrategy,
    TTableSchemaColumns,
)
from dlt.common.wei import EVM_DECIMAL_PRECISION

TLoaderParallelismStrategy = Literal["parallel", "table-sequential", "sequential"]
LOADER_FILE_FORMATS: Set[TLoaderFileFormat] = set(get_args(TLoaderFileFormat))


class LoaderFileFormatSelector(Protocol):
    """Selects preferred and supported file formats for a given table schema"""

    @staticmethod
    def __call__(
        preferred_loader_file_format: TLoaderFileFormat,
        supported_loader_file_formats: Sequence[TLoaderFileFormat],
        /,
        *,
        table_schema: TTableSchema,
    ) -> Tuple[TLoaderFileFormat, Sequence[TLoaderFileFormat]]: ...


class MergeStrategySelector(Protocol):
    """Selects right set of merge strategies for a given table schema"""

    @staticmethod
    def __call__(
        supported_merge_strategies: Sequence[TLoaderMergeStrategy],
        /,
        *,
        table_schema: TTableSchema,
    ) -> Sequence["TLoaderMergeStrategy"]: ...


class ReplaceStrategySelector(Protocol):
    """Selects right set of replace strategies for a given table schema"""

    @staticmethod
    def __call__(
        supported_replace_strategies: Sequence[TLoaderReplaceStrategy],
        /,
        *,
        table_schema: TTableSchema,
    ) -> Sequence["TLoaderReplaceStrategy"]: ...


class DataTypeMapper(ABC):
    def __init__(self, capabilities: "DestinationCapabilitiesContext") -> None:
        """Maps dlt data types into destination data types"""
        self.capabilities = capabilities

    @abstractmethod
    def to_destination_type(self, column: TColumnSchema, table: PreparedTableSchema) -> str:
        """Gets destination data type for a particular `column` in prepared `table`"""
        pass

    @abstractmethod
    def from_destination_type(
        self, db_type: str, precision: Optional[int], scale: Optional[int]
    ) -> TColumnType:
        """Gets column type from db type"""
        pass

    @abstractmethod
    def ensure_supported_type(
        self,
        column: TColumnSchema,
        table: PreparedTableSchema,
        loader_file_format: TLoaderFileFormat,
    ) -> None:
        """Makes sure that dlt type in `column` in prepared `table`  is supported by the destination for a given file format"""
        pass


class UnsupportedTypeMapper(DataTypeMapper):
    """Type Mapper that can't map any type"""

    def to_destination_type(self, column: TColumnSchema, table: PreparedTableSchema) -> str:
        raise NotImplementedError("No types are supported, use real type mapper")

    def from_destination_type(
        self, db_type: str, precision: Optional[int], scale: Optional[int]
    ) -> TColumnType:
        raise NotImplementedError("No types are supported, use real type mapper")

    def ensure_supported_type(
        self,
        column: TColumnSchema,
        table: PreparedTableSchema,
        loader_file_format: TLoaderFileFormat,
    ) -> None:
        raise TerminalValueError(
            "No types are supported, use real type mapper", column["data_type"]
        )


TCasefoldIdentifier = Callable[[str], str]


@configspec
class DestinationCapabilitiesContext(ContainerInjectableContext):
    """Injectable destination capabilities required for many Pipeline stages ie. normalize"""

    # do not allow to create default value, destination caps must be always explicitly inserted into container
    can_create_default: ClassVar[bool] = False

    preferred_loader_file_format: TLoaderFileFormat = None
    supported_loader_file_formats: Sequence[TLoaderFileFormat] = None
    loader_file_format_selector: LoaderFileFormatSelector = None
    """Callable that adapts `preferred_loader_file_format` and `supported_loader_file_formats` at runtime."""
    preferred_table_format: TTableFormat = None
    supported_table_formats: Sequence[TTableFormat] = None
    type_mapper: Optional[Type[DataTypeMapper]] = None
    recommended_file_size: Optional[int] = None
    """Recommended file size in bytes when writing extract/load files"""
    preferred_staging_file_format: Optional[TLoaderFileFormat] = None
    supported_staging_file_formats: Sequence[TLoaderFileFormat] = None
    format_datetime_literal: Callable[..., str] = None
    escape_identifier: Callable[[str], str] = None
    "Escapes table name, column name and other identifiers"
    escape_literal: Callable[[Any], Any] = None
    "Escapes string literal"
    casefold_identifier: TCasefoldIdentifier = str
    """Casing function applied by destination to represent case insensitive identifiers."""
    has_case_sensitive_identifiers: bool = None
    """Tells if destination supports case sensitive identifiers"""
    decimal_precision: Tuple[int, int] = None
    wei_precision: Tuple[int, int] = None
    max_identifier_length: int = None
    max_column_identifier_length: int = None
    max_query_length: int = None
    is_max_query_length_in_bytes: bool = None
    max_text_data_type_length: int = None
    is_max_text_data_type_length_in_bytes: bool = None
    supports_transactions: bool = None
    supports_ddl_transactions: bool = None
    # use naming convention in the schema
    naming_convention: TNamingConventionReferenceArg = None
    alter_add_multi_column: bool = True
    supports_create_table_if_not_exists: bool = True
    supports_truncate_command: bool = True
    schema_supports_numeric_precision: bool = True
    timestamp_precision: int = 6
    """Default precision of the timestamp type"""
    max_timestamp_precision: int = 6
    """Maximum supported timestamp precision"""

    max_rows_per_insert: Optional[int] = None
    insert_values_writer_type: str = "default"
    supports_multiple_statements: bool = True
    supports_clone_table: bool = False
    """Destination supports CREATE TABLE ... CLONE ... statements"""

    max_table_nesting: Optional[int] = None
    """Allows a destination to overwrite max_table_nesting from source"""

    supported_merge_strategies: Sequence[TLoaderMergeStrategy] = None
    merge_strategies_selector: MergeStrategySelector = None
    supported_replace_strategies: Sequence[TLoaderReplaceStrategy] = None
    replace_strategies_selector: ReplaceStrategySelector = None

    max_parallel_load_jobs: Optional[int] = None
    """The destination can set the maximum amount of parallel load jobs being executed"""
    loader_parallelism_strategy: Optional[TLoaderParallelismStrategy] = None
    """The destination can override the parallelism strategy"""

    max_query_parameters: Optional[int] = None
    """The maximum number of parameters that can be supplied in a single parametrized query"""

    supports_native_boolean: bool = True
    """The destination supports a native boolean type, otherwise bool columns are usually stored as integers"""

    supports_nested_types: bool = False
    """Tells if destination can write nested types, currently only destinations storing parquet are supported"""

    enforces_nulls_on_alter: bool = True
    """Tells if destination enforces null constraints when adding NOT NULL columns to existing tables"""

    sqlglot_dialect: Optional[TSqlGlotDialect] = None
    """The SQL dialect used by sqlglot to transpile a query to match the destination syntax."""

    parquet_format: Optional[ParquetFormatConfiguration] = None
    """Parquet format preferred by this destination"""

    supports_tz_aware_datetime: bool = True
    """The destination can store datetime with timezone"""

    supports_naive_datetime: bool = True
    """The destination can store datetime without timezone"""

    dialect_capabilities: Optional[Any] = None
    """Capabilities of selected dialect for multi-dialect destinations ie. SQLAlchemy"""

    def generates_case_sensitive_identifiers(self) -> bool:
        """Tells if capabilities as currently adjusted, will generate case sensitive identifiers"""
        # must have case sensitive support and folding function must preserve casing
        return self.has_case_sensitive_identifiers and self.casefold_identifier is str

    @staticmethod
    def generic_capabilities(
        preferred_loader_file_format: TLoaderFileFormat = None,
        naming_convention: TNamingConventionReferenceArg = None,
        loader_file_format_selector: LoaderFileFormatSelector = None,
        supported_table_formats: Sequence[TTableFormat] = None,
        supported_merge_strategies: Sequence[TLoaderMergeStrategy] = None,
        merge_strategies_selector: MergeStrategySelector = None,
    ) -> "DestinationCapabilitiesContext":
        from dlt.common.data_writers.escape import format_datetime_literal

        caps = DestinationCapabilitiesContext()
        caps.preferred_loader_file_format = preferred_loader_file_format
        caps.supported_loader_file_formats = ["jsonl", "insert_values", "parquet", "csv", "model"]
        caps.loader_file_format_selector = loader_file_format_selector
        caps.preferred_staging_file_format = None
        caps.supported_staging_file_formats = []
        caps.naming_convention = naming_convention or caps.naming_convention
        caps.escape_identifier = str
        caps.supported_table_formats = supported_table_formats or []
        caps.escape_literal = serialize_value
        caps.casefold_identifier = str
        caps.has_case_sensitive_identifiers = True
        caps.format_datetime_literal = format_datetime_literal
        caps.decimal_precision = (DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE)
        caps.wei_precision = (EVM_DECIMAL_PRECISION, 0)
        caps.max_identifier_length = 65536
        caps.max_column_identifier_length = 65536
        caps.max_query_length = 32 * 1024 * 1024
        caps.is_max_query_length_in_bytes = True
        caps.max_text_data_type_length = 1024 * 1024 * 1024
        caps.is_max_text_data_type_length_in_bytes = True
        caps.supports_ddl_transactions = True
        caps.supports_transactions = True
        caps.supports_multiple_statements = True
        caps.supported_merge_strategies = supported_merge_strategies or []
        caps.merge_strategies_selector = merge_strategies_selector
        return caps

    def get_type_mapper(self, *args: Any, **kwargs: Any) -> DataTypeMapper:
        return self.type_mapper(self, *args, **kwargs)


def merge_caps_file_formats(
    destination: str,
    staging: str,
    dest_caps: DestinationCapabilitiesContext,
    stage_caps: DestinationCapabilitiesContext,
) -> Tuple[TLoaderFileFormat, Sequence[TLoaderFileFormat]]:
    """Merges preferred and supported file formats from destination and staging.
    Returns new preferred file format and all possible formats.
    """
    possible_file_formats = dest_caps.supported_loader_file_formats
    if stage_caps:
        if not dest_caps.supported_staging_file_formats:
            raise DestinationLoadingViaStagingNotSupported(destination)
        possible_file_formats = [
            f
            for f in dest_caps.supported_staging_file_formats
            if f in stage_caps.supported_loader_file_formats
        ]
    if len(possible_file_formats) == 0:
        raise DestinationIncompatibleLoaderFileFormatException(
            destination, staging, None, possible_file_formats
        )
    if not stage_caps:
        if not dest_caps.preferred_loader_file_format:
            raise DestinationLoadingWithoutStagingNotSupported(destination)
        requested_file_format = dest_caps.preferred_loader_file_format
    elif stage_caps and dest_caps.preferred_staging_file_format in possible_file_formats:
        requested_file_format = dest_caps.preferred_staging_file_format
    else:
        requested_file_format = possible_file_formats[0] if len(possible_file_formats) > 0 else None
    return requested_file_format, possible_file_formats


def adjust_column_schema_to_capabilities(
    column: TColumnSchema, caps: DestinationCapabilitiesContext
) -> TColumnSchema:
    """Modify column schema in place according to destination capabilities.
    * timestamp and time precision are limited by min(destination max, parquet writer max)
    * timezone is always removed (enables default behavior)
    * default timestamp precision for destination is removed
    """
    data_type = column.get("data_type")
    if data_type in ("timestamp", "time"):
        precision = column.get("precision", caps.timestamp_precision)
        # limit precision according to caps: min of parquet writer and destination max
        precision = min(
            precision,
            caps.max_timestamp_precision,
            (
                caps.parquet_format.max_timestamp_precision()
                if caps.parquet_format
                else ParquetFormatConfiguration().max_timestamp_precision()
            ),
        )

        # remove timezone flag to fallback to default tz-aware and let normalizer to correct the data
        column.pop("timezone", None)

        # do not write default precision
        if precision == caps.timestamp_precision:
            column.pop("precision", None)
        elif column.get("precision") != precision:
            column["precision"] = precision

    if data_type == "json" and not caps.supports_nested_types:
        column.pop("x-nested-type", None)  # type: ignore[typeddict-item]

    return column


def adjust_schema_to_capabilities(
    columns: TTableSchemaColumns, caps: DestinationCapabilitiesContext
) -> TTableSchemaColumns:
    """Modifies `columns` schema in place according to destination `caps`. See `adjust_column_schema_to_capabilities`
    for a list of adjustments.
    NOTE: that function is intended to be used on schemas inferred from the data,
    before user explicit hints are applied as it overrides ie. timezone and precision hints.

    Returns modified columns schema
    """
    for col in columns.values():
        adjust_column_schema_to_capabilities(col, caps)
    return columns



from dlt.extract.resource import DltResource, with_table_name, with_hints
from dlt.extract.hints import make_hints, make_nested_hints
from dlt.extract.source import DltSource
from dlt.extract.reference import SourceFactory, SourceReference
from dlt.extract.decorators import source, resource, transformer, defer
from dlt.extract.incremental import Incremental
from dlt.extract.wrappers import wrap_additional_type
from dlt.extract.extractors import materialize_schema_item, with_file_import

__all__ = [
    "DltResource",
    "DltSource",
    "SourceFactory",
    "SourceReference",
    "with_table_name",
    "with_hints",
    "with_file_import",
    "make_hints",
    "make_nested_hints",
    "source",
    "resource",
    "transformer",
    "defer",
    "Incremental",
    "wrap_additional_type",
    "materialize_schema_item",
]



from __future__ import annotations as _annotations
import collections.abc
from copy import copy
from typing import (
    Dict,
    Generic,
    Optional,
    Set,
    List,
    Tuple,
    Type,
    Union,
    Any,
)

from dlt.common.data_types import py_type_to_sc_type
from dlt.common.exceptions import MissingDependencyException
from dlt.common.schema import DataValidationError
from dlt.common.schema.typing import TSchemaEvolutionMode, TTableSchemaColumns
from dlt.common.normalizers.naming.snake_case import NamingConvention as SnakeCaseNamingConvention
from dlt.common.typing import (
    TypedDict,
    Annotated,
    get_args,
    get_origin,
    TypeVar,
    TDataItem,
    TDataItems,
    extract_union_types,
    is_annotated,
    is_optional_type,
    extract_inner_type,
    is_list_generic_type,
    is_dict_generic_type,
    is_subclass,
    is_union_type,
)
import warnings

from dlt.common.warnings import Dlt100DeprecationWarning

try:
    from pydantic import VERSION as PYDANTIC_VERSION

    if not PYDANTIC_VERSION.startswith("2."):
        raise ImportError(f"Found pydantic {PYDANTIC_VERSION} but dlt requires pydantic>=2.0")
    from pydantic import BaseModel, ValidationError, Json, create_model
    from pydantic.fields import FieldInfo
    from pydantic.warnings import PydanticDeprecationWarning

    warnings.filterwarnings("ignore", category=PydanticDeprecationWarning)
except ImportError:
    raise MissingDependencyException(
        "dlt Pydantic helpers", ["pydantic>=2.0"], "Pydantic 2.x is required"
    )

_TPydanticModel = TypeVar("_TPydanticModel", bound=BaseModel)


snake_case_naming_convention = SnakeCaseNamingConvention()


def _is_set_origin(origin: type) -> bool:
    """Check if a type origin is set-like (set or frozenset)."""
    try:
        return issubclass(origin, collections.abc.Set)
    except TypeError:
        return False


class ListModel(BaseModel, Generic[_TPydanticModel]):
    items: List[_TPydanticModel]


class DltConfig(TypedDict, total=False):
    """dlt configuration that can be attached to a Pydantic model.

    Example:
        >>> class ItemModel(BaseModel):
        >>>     field: int
        >>>     dlt_config: ClassVar[DltConfig] = {"return_validated_models": True}

    Options:
        skip_nested_types:
            If True, columns of complex types (`dict`, `list`, `BaseModel`) are excluded
            from the schema generated from the model.

        skip_complex_types:  # deprecated

        return_validated_models:
            If True, the Pydantic validator returns validated Pydantic model instances
            instead of converting them to dictionaries during extraction/transform steps.
            Defaults to False to preserve current behavior.

        is_authoritative_model:
            If True, the Pydantic model is treated as the authoritative source of table
            schema. Columns derived from the model are pre-merged into the schema before
            contract checks, so they bypass column/data_type contract enforcement.
            When None (default), dlt automatically determines whether to treat the model
            as authoritative.
    """

    skip_nested_types: bool
    skip_complex_types: bool  # deprecated
    return_validated_models: bool
    is_authoritative_model: Optional[bool]


def _build_discriminator_map(
    model: Type[BaseModel],
) -> Optional[Tuple[str, Dict[str, Type[BaseModel]]]]:
    """For a RootModel with a discriminated union, extract the discriminator field name
    and a mapping from literal discriminator values to their variant model classes.

    Returns None if the model is not a RootModel or has no string discriminator.
    """
    if not getattr(model, "__pydantic_root_model__", False):
        return None
    root_field = model.model_fields.get("root")
    if not root_field:
        return None

    ann = root_field.annotation
    discriminator: Optional[str] = None

    if is_annotated(ann):
        args = get_args(ann)
        # metadata may be FieldInfo directly or wrapped in a tuple by _process_annotation
        for a in args[1:]:
            items = a if isinstance(a, (list, tuple)) else (a,)
            for item in items:
                if isinstance(item, FieldInfo) and isinstance(item.discriminator, str):
                    discriminator = item.discriminator
                    break
            if discriminator:
                break
        union_args = get_args(args[0])
    else:
        return None

    if not discriminator or not union_args:
        return None

    mapping: Dict[str, Type[BaseModel]] = {}
    for member in union_args:
        member_field = member.model_fields.get(discriminator)
        if member_field:
            for lit_val in get_args(member_field.annotation):
                mapping[str(lit_val)] = member
    return discriminator, mapping


def resolve_variant_model(
    model: Type[BaseModel],
    item: Any,
    discriminator_map: Optional[Tuple[str, Dict[str, Type[BaseModel]]]] = None,
) -> Optional[Type[BaseModel]]:
    """Resolve a discriminated union variant from an item's discriminator value.

    Args:
        model: A pydantic model, typically a RootModel with a discriminated union.
        item: A dict or model instance to read the discriminator value from.
        discriminator_map: Pre-computed result of `_build_discriminator_map`. Computed
            from `model` when not provided.
    """
    if discriminator_map is None:
        discriminator_map = _build_discriminator_map(model)
    if discriminator_map is None:
        return model
    disc_field, mapping = discriminator_map
    disc_value = item.get(disc_field) if isinstance(item, dict) else getattr(item, disc_field, None)
    if disc_value is None:
        return None
    return mapping.get(str(disc_value))


def pydantic_to_table_schema_columns(
    model: Union[BaseModel, Type[BaseModel]],
) -> TTableSchemaColumns:
    """Convert a pydantic model to a table schema columns dict

    See also DltConfig for more control over how the schema is created

    Args:
        model: The pydantic model to convert. Can be a class or an instance.

    Returns:
        TTableSchemaColumns: table schema columns dict
    """
    # for RootModel with discriminated union, return columns common to all variants
    if getattr(model, "__pydantic_root_model__", False):
        disc_map = _build_discriminator_map(model)  # type: ignore[arg-type]
        if disc_map is not None:
            variants = list(disc_map[1].values())
            all_cols = [pydantic_to_table_schema_columns(v) for v in variants]
            common_keys = set.intersection(*(set(c.keys()) for c in all_cols))
            return {k: v for k, v in all_cols[0].items() if k in common_keys}

    skip_nested_types = False
    if hasattr(model, "dlt_config"):
        if "skip_complex_types" in model.dlt_config:
            warnings.warn(
                "`skip_complex_types` is deprecated, use `skip_nested_types` instead.",
                Dlt100DeprecationWarning,
                stacklevel=2,
            )
            skip_nested_types = model.dlt_config["skip_complex_types"]
        else:
            skip_nested_types = model.dlt_config.get("skip_nested_types", False)

    result: TTableSchemaColumns = {}
    model_cls: Type[BaseModel] = model if isinstance(model, type) else type(model)

    for field_name, field in model_cls.model_fields.items():
        annotation = field.annotation
        if inner_annotation := getattr(annotation, "inner_type", None):
            # This applies to pydantic.Json fields, the inner type is the type after json parsing
            # (In pydantic 2 the outer annotation is the final type)
            annotation = inner_annotation

        nullable = is_optional_type(annotation)

        inner_type = extract_inner_type(annotation)
        if is_union_type(inner_type):
            # TODO: order those types deterministically before getting first one
            # order of the types in union is in many cases not deterministic
            # https://docs.python.org/3/library/typing.html#typing.get_args
            first_argument_type = get_args(inner_type)[0]
            inner_type = extract_inner_type(first_argument_type)

        if inner_type is Json:  # Same as `field: Json[Any]`
            inner_type = Any  # type: ignore[assignment]

        if inner_type is Any:  # Any fields will be inferred from data
            continue

        if is_list_generic_type(inner_type):
            inner_type = list
        elif is_dict_generic_type(inner_type):
            inner_type = dict

        is_inner_type_pydantic_model = False
        name = field.alias or field_name
        try:
            data_type = py_type_to_sc_type(inner_type)
        except TypeError:
            if is_subclass(inner_type, BaseModel):
                data_type = "json"
                is_inner_type_pydantic_model = True
            else:
                # try to coerce unknown type to text
                data_type = "text"

        if is_inner_type_pydantic_model and not skip_nested_types:
            result[name] = {
                "name": name,
                "data_type": "json",
                "nullable": nullable,
            }
        elif is_inner_type_pydantic_model:
            # This case is for a single field schema/model
            # we need to generate snake_case field names
            # and return flattened field schemas
            schema_hints = pydantic_to_table_schema_columns(inner_type)

            for field_name, hints in schema_hints.items():
                schema_key = snake_case_naming_convention.make_path(name, field_name)
                result[schema_key] = {
                    **hints,
                    "name": snake_case_naming_convention.make_path(name, hints["name"]),
                    # if the outer field containing the nested mode is optional,
                    # then each field in the model itself has to be nullable as well,
                    # as otherwise we end up with flattened non-nullable optional nested fields
                    "nullable": hints["nullable"] or nullable,
                }
        elif data_type == "json" and skip_nested_types:
            continue
        else:
            result[name] = {
                "name": name,
                "data_type": data_type,
                "nullable": nullable,
            }

    return result


def column_mode_to_extra(column_mode: TSchemaEvolutionMode) -> str:
    extra = "forbid"
    if column_mode == "evolve":
        extra = "allow"
    elif column_mode == "discard_value":
        extra = "ignore"
    return extra


def extra_to_column_mode(extra: str) -> TSchemaEvolutionMode:
    if extra == "forbid":
        return "freeze"
    if extra == "allow":
        return "evolve"
    return "discard_value"


def get_extra_from_model(model: Type[BaseModel]) -> Optional[str]:
    """Returns the extra setting from the model or None if not explicitly set."""
    return model.model_config.get("extra")


def apply_schema_contract_to_model(
    model: Type[_TPydanticModel],
    column_mode: TSchemaEvolutionMode,
    data_mode: TSchemaEvolutionMode = "freeze",
    _child_models: Optional[Dict[int, Type[BaseModel]]] = None,
) -> Type[_TPydanticModel]:
    """Configures or re-creates `model` so it behaves according to `column_mode` and `data_mode` settings.

    `column_mode` sets the model behavior when unknown field is found.
    `data_mode` sets model behavior when known field does not validate. currently `evolve` and `freeze` are supported here.

    `discard_row` is implemented in `validate_item`.
    """
    # save dlt_config before model may be replaced by the evolve path
    original_dlt_config = getattr(model, "dlt_config", None)

    if data_mode == "evolve":
        # create a lenient model that accepts any data
        if getattr(model, "__pydantic_root_model__", False):
            from pydantic import RootModel as PydanticRootModel

            model = type(
                model.__name__ + "Any",
                (PydanticRootModel[Any],),
                {"__module__": model.__module__},
            )
        else:
            model = create_model(model.__name__ + "Any", **{n: (Any, None) for n in model.model_fields})  # type: ignore
    elif data_mode == "discard_value":
        raise NotImplementedError(
            "`data_mode='discard_value'`. Cannot discard defined fields with validation errors"
            " using Pydantic models."
        )

    extra = column_mode_to_extra(column_mode)

    if extra == (get_extra_from_model(model) or "ignore"):
        # no need to change the model, but restore dlt_config lost by the evolve path
        if original_dlt_config:
            model.dlt_config = original_dlt_config  # type: ignore[attr-defined]
        return model

    config = copy(model.model_config)
    config["extra"] = extra  # type: ignore[typeddict-item]

    if _child_models is None:
        _child_models = {}

    def _process_annotation(t_: Type[Any]) -> Type[Any]:
        """Recursively recreates models with applied schema contract"""
        if is_annotated(t_):
            a_t, *a_m = get_args(t_)
            return Annotated[_process_annotation(a_t), tuple(a_m)]  # type: ignore[return-value]
        origin = get_origin(t_)
        # tuple must be checked before is_list_generic_type (tuple is a Sequence)
        if origin is tuple:
            args = get_args(t_)
            if not args:
                return t_
            if len(args) == 2 and args[1] is Ellipsis:
                # variable-length: Tuple[T, ...]
                return origin[_process_annotation(args[0]), ...]  # type: ignore[no-any-return]
            # heterogeneous: Tuple[T1, T2, ...]
            processed = tuple(_process_annotation(a) for a in args)
            return origin[processed]  # type: ignore[no-any-return]
        elif is_list_generic_type(t_):
            l_t: Type[Any] = get_args(t_)[0]
            return origin[_process_annotation(l_t)]  # type: ignore[no-any-return]
        elif is_dict_generic_type(t_):
            k_t: Type[Any]
            v_t: Type[Any]
            k_t, v_t = get_args(t_)
            return origin[k_t, _process_annotation(v_t)]  # type: ignore[no-any-return]
        # set/frozenset are not Sequence or Mapping so need a separate check
        elif origin is not None and _is_set_origin(origin):
            s_t: Type[Any] = get_args(t_)[0]
            return origin[_process_annotation(s_t)]  # type: ignore[no-any-return]
        elif is_union_type(t_):
            u_t_s = tuple(_process_annotation(u_t) for u_t in extract_union_types(t_))
            return Union[u_t_s]  # type: ignore[return-value]
        elif is_subclass(t_, BaseModel):
            # types must be same before and after processing
            if id(t_) in _child_models:
                return _child_models[id(t_)]
            else:
                _child_models[id(t_)] = child_model = apply_schema_contract_to_model(
                    t_, column_mode, data_mode, _child_models=_child_models
                )
                return child_model
        return t_

    def _rebuild_annotated(f: Any) -> Type[Any]:
        if hasattr(f, "rebuild_annotation"):
            return f.rebuild_annotation()  # type: ignore[no-any-return]
        else:
            return f.annotation  # type: ignore[no-any-return]

    if getattr(model, "__pydantic_root_model__", False):
        from pydantic import RootModel as PydanticRootModel

        root_field = model.model_fields.get("root")
        if root_field:
            processed_ann = _process_annotation(_rebuild_annotated(root_field))
            new_rm = type(
                model.__name__ + "Extra" + extra.title(),
                (PydanticRootModel[processed_ann],),  # type: ignore[valid-type]
                {"__module__": model.__module__},
            )
            if original_dlt_config:
                new_rm.dlt_config = original_dlt_config  # type: ignore[attr-defined]
            return new_rm

    processed_fields = {
        n: (_process_annotation(_rebuild_annotated(f)), f) for n, f in model.model_fields.items()
    }

    # use __base__ to inherit validators (@field_validator, @model_validator)
    new_model: Type[_TPydanticModel] = create_model(  # type: ignore[call-overload]
        model.__name__ + "Extra" + extra.title(),
        __base__=model,
        **processed_fields,
    )
    # override config on the subclass and rebuild to pick up the new extra setting
    new_model.model_config = config
    new_model.model_rebuild(force=True)

    # restore dlt_config: the evolve path replaces model with a bare "Any" model that
    # lacks dlt_config. For the non-evolve path __base__ inheritance preserves it,
    # but setting it explicitly is harmless and keeps both paths consistent.
    if original_dlt_config:
        new_model.dlt_config = original_dlt_config  # type: ignore[attr-defined]
    return new_model


def create_list_model(
    model: Type[_TPydanticModel],
    column_mode: TSchemaEvolutionMode = "freeze",
    data_mode: TSchemaEvolutionMode = "freeze",
) -> Type[ListModel[_TPydanticModel]]:
    """Creates a model from `model` for validating list of items in batch.

    When `column_mode` or `data_mode` is `discard_row`, creates a lenient list model
    that uses WrapValidator to turn invalid items into None (filtered by the caller).
    Otherwise creates a strict list model for batch validation.
    """
    if column_mode == "discard_row" or data_mode == "discard_row":
        from pydantic.functional_validators import WrapValidator

        def _lenient_item_validator(value: Any, handler: Any) -> Optional[_TPydanticModel]:
            try:
                return handler(value)  # type: ignore[no-any-return]
            except ValidationError as val_err:
                # re-raise model_type errors (non-mapping items) — cannot discard those
                for err in val_err.errors():
                    if err["type"] == "model_type":
                        raise
                return None
            except Exception:
                return None

        item_type = Annotated[Optional[model], WrapValidator(_lenient_item_validator)]  # type: ignore[valid-type]
        return create_model(
            "LenientList" + model.__name__,
            items=(List[item_type], ...),  # type: ignore[return-value]
        )

    return create_model(
        "List" + model.__name__,
        items=(List[model], ...),  # type: ignore[return-value,valid-type]
    )


def _classify_validation_errors(
    table_name: str,
    model: Type[BaseModel],
    item: TDataItem,
    exc: ValidationError,
    column_mode: TSchemaEvolutionMode,
    data_mode: TSchemaEvolutionMode,
) -> None:
    """Classifies validation errors and raises DataValidationError for freeze mode.

    For discard_row mode, returns without raising so the caller can discard the item.
    For model_type errors (item is not a mapping), always re-raises.
    """
    for err in exc.errors():
        if err["type"] == "model_type":
            raise exc
        if err["type"] == "extra_forbidden":
            if column_mode == "freeze":
                raise DataValidationError(
                    None,
                    table_name,
                    str(err["loc"]),
                    "columns",
                    "freeze",
                    model,
                    {"columns": "freeze"},
                    item,
                    err["msg"],
                ) from exc
            elif column_mode == "discard_row":
                return
            raise NotImplementedError(f"`{column_mode=:}` not implemented for Pydantic validation")
        else:
            if data_mode == "freeze":
                raise DataValidationError(
                    None,
                    table_name,
                    str(err["loc"]),
                    "data_type",
                    "freeze",
                    model,
                    {"data_type": "freeze"},
                    item,
                    err["msg"],
                ) from exc
            elif data_mode == "discard_row":
                return
            raise NotImplementedError(f"`{data_mode=:}` not implemented for Pydantic validation")


def validate_and_filter_items(
    table_name: str,
    list_model: Type[ListModel[_TPydanticModel]],
    items: List[TDataItem],
    column_mode: TSchemaEvolutionMode,
    data_mode: TSchemaEvolutionMode,
) -> List[_TPydanticModel]:
    """Validates list of `item` with `list_model` and returns parsed Pydantic models."""
    if column_mode == "discard_row" or data_mode == "discard_row":
        # lenient path: WrapValidator already converted bad items to None
        result = list_model(items=items).items
        return [item for item in result if item is not None]

    # strict path: batch validate, classify errors on failure
    try:
        return list_model(items=items).items
    except ValidationError as e:
        for err in e.errors():
            if err["type"] == "model_type":
                raise
            # loc is ("items", <index>, <field>) for item-level errors
            if len(err["loc"]) < 2:
                # list-level error (e.g. items not a list) — re-raise as-is
                raise
            err_idx = int(err["loc"][1])
            err_item = items[err_idx]
            _classify_validation_errors(table_name, list_model, err_item, e, column_mode, data_mode)
        raise AssertionError("unreachable")


def validate_and_filter_item(
    table_name: str,
    model: Type[_TPydanticModel],
    item: TDataItems,
    column_mode: TSchemaEvolutionMode,
    data_mode: TSchemaEvolutionMode,
) -> Optional[_TPydanticModel]:
    """Validates `item` against model `model` and returns an instance of it.

    Returns None for discard_row when validation fails, raises DataValidationError
    for freeze mode.
    """
    try:
        return model.model_validate(item)
    except ValidationError as e:
        _classify_validation_errors(table_name, model, item, e, column_mode, data_mode)
        return None



import time
from typing import Any, Dict, Optional, Union

import dlt
from dlt.common.configuration import known_sections, with_config
from dlt.helpers.dbt_cloud.client import DBTCloudClientV2
from dlt.helpers.dbt_cloud.configuration import DBTCloudConfiguration


@with_config(
    spec=DBTCloudConfiguration,
    sections=(known_sections.DBT_CLOUD,),
)
def run_dbt_cloud_job(
    credentials: DBTCloudConfiguration = dlt.secrets.value,
    job_id: Union[int, str, None] = None,
    data: Optional[Dict[Any, Any]] = None,
    wait_for_outcome: bool = True,
    wait_seconds: int = 10,
) -> Dict[Any, Any]:
    """
    Trigger a dbt Cloud job run and retrieve its status.

    Args:
        credentials (DBTCloudConfiguration): Configuration parameters for dbt Cloud.
            Defaults to dlt.secrets.value.
        job_id (int | str, optional): The ID of the specific job to run.
            If not provided, it will use the job ID specified in the credentials.
            Defaults to None.
        data (dict, optional): Additional data to include when triggering the job run.
            Defaults to None.
            Fields of data:
                 '{
                    "cause": "string",
                    "git_sha": "string",
                    "git_branch": "string",
                    "azure_pull_request_id": integer,
                    "github_pull_request_id": integer,
                    "gitlab_merge_request_id": integer,
                    "schema_override": "string",
                    "dbt_version_override": "string",
                    "threads_override": integer,
                    "target_name_override": "string",
                    "generate_docs_override": boolean,
                    "timeout_seconds_override": integer,
                    "steps_override": [
                        "string"
                    ]
                }'
        wait_for_outcome (bool, optional): Whether to wait for the job run to complete before returning.
            Defaults to True.
        wait_seconds (int, optional): The interval (in seconds) between status checks while waiting for completion.
            Defaults to 10.

    Returns:
        dict: A dictionary containing the status information of the job run.

    Raises:
        InvalidCredentialsException: If account_id or job_id is missing.
    """
    operator = DBTCloudClientV2(
        api_token=credentials.api_token,
        account_id=credentials.account_id,
    )
    json_data = {"cause": credentials.cause}

    if credentials.git_sha:
        json_data["git_sha"] = credentials.git_sha

    elif credentials.git_branch:
        json_data["git_branch"] = credentials.git_branch

    if credentials.schema_override:
        json_data["schema_override"] = credentials.schema_override

    if data:
        json_data.update(json_data)

    job_id = job_id or credentials.job_id

    run_id = operator.trigger_job_run(job_id=job_id, data=json_data)
    status = operator.get_run_status(run_id)

    if wait_for_outcome and status["in_progress"]:
        while True:
            status = operator.get_run_status(run_id)
            if not status["in_progress"]:
                break

            time.sleep(wait_seconds)

    return status


@with_config(
    spec=DBTCloudConfiguration,
    sections=(known_sections.DBT_CLOUD,),
)
def get_dbt_cloud_run_status(
    credentials: DBTCloudConfiguration = dlt.secrets.value,
    run_id: Union[int, str, None] = None,
    wait_for_outcome: bool = True,
    wait_seconds: int = 10,
) -> Dict[Any, Any]:
    """
    Retrieve the status of a dbt Cloud job run.

    Args:
        credentials (DBTCloudConfiguration): Configuration parameters for dbt Cloud.
            Defaults to dlt.secrets.value.
        run_id (int | str, optional): The ID of the specific job run to retrieve status for.
            If not provided, it will use the run ID specified in the credentials.
            Defaults to None.
        wait_for_outcome (bool, optional): Whether to wait for the job run to complete before returning.
            Defaults to True.
        wait_seconds (int, optional): The interval (in seconds) between status checks while waiting for completion.
            Defaults to 10.

    Returns:
        dict: A dictionary containing the status information of the specified job run.

    Raises:
        InvalidCredentialsException: If account_id or run_id is missing.
    """
    operator = DBTCloudClientV2(
        api_token=credentials.api_token,
        account_id=credentials.account_id,
    )
    run_id = run_id or credentials.run_id
    status = operator.get_run_status(run_id)

    if wait_for_outcome and status["in_progress"]:
        while True:
            status = operator.get_run_status(run_id)
            if not status["in_progress"]:
                break

            time.sleep(wait_seconds)

    return status



import dataclasses
from typing import ClassVar, Dict, List, Any, Final, cast, Optional
from typing_extensions import Annotated

from dlt.common.configuration import configspec
from dlt.common.configuration.specs import ConnectionStringCredentials
from dlt.common.configuration.specs.base_configuration import NotResolved
from dlt.common.destination.client import (
    DestinationClientDwhWithStagingConfiguration,
)
from dlt.common.utils import digest128
from dlt.destinations.impl.clickhouse.typing import TSecureConnection, TTableEngineType


@configspec(init=False)
class ClickHouseCredentials(ConnectionStringCredentials):
    drivername: str = "clickhouse"
    host: str = None
    """Host with running ClickHouse server."""
    port: int = 9440
    """Native port ClickHouse server is bound to. Defaults to 9440."""
    http_port: int = 8443
    """HTTP Port to connect to ClickHouse server's HTTP interface.
    The HTTP port is additionaly needed for loading files without a staging destination.
    Defaults to 8443."""
    username: str = "default"
    """Database user. Defaults to 'default'."""
    database: str = "default"
    """database connect to. Defaults to 'default'."""
    secure: TSecureConnection = 1
    """Enables TLS encryption when connecting to ClickHouse Server. 0 means no encryption, 1 means encrypted."""
    connect_timeout: int = 15
    """Timeout for establishing connection. Defaults to 10 seconds."""
    send_receive_timeout: int = 300
    """Timeout for sending and receiving data. Defaults to 300 seconds."""
    s3_extra_credentials: Optional[Dict[str, str]] = None
    """Arguments to pass to the `extra_credentials` function specifically for S3 function."""

    __config_gen_annotations__: ClassVar[List[str]] = [
        "host",
        "port",
        "http_port",
        "database",
        "username",
        "password",
        "s3_extra_credentials",
    ]

    def parse_native_representation(self, native_value: Any) -> None:
        super().parse_native_representation(native_value)
        self.connect_timeout = int(self.query.get("connect_timeout", self.connect_timeout))
        self.send_receive_timeout = int(
            self.query.get("send_receive_timeout", self.send_receive_timeout)
        )
        self.secure = cast(TSecureConnection, int(self.query.get("secure", self.secure)))

    def get_query(self) -> Dict[str, Any]:
        query = dict(super().get_query())
        query.update(
            {
                "connect_timeout": str(self.connect_timeout),
                "send_receive_timeout": str(self.send_receive_timeout),
                "secure": 1 if self.secure else 0,
                "allow_experimental_lightweight_delete": 1,
                "enable_http_compression": 1,
                "date_time_input_format": "best_effort",
            }
        )
        return query


@configspec
class ClickHouseClientConfiguration(DestinationClientDwhWithStagingConfiguration):
    destination_type: Final[str] = dataclasses.field(  # type: ignore[misc]
        default="clickhouse", init=False, repr=False, compare=False
    )
    # allow empty dataset names
    dataset_name: Annotated[Optional[str], NotResolved()] = dataclasses.field(
        default=None, init=False, repr=False, compare=False
    )
    credentials: ClickHouseCredentials = None

    dataset_table_separator: str = "___"
    """Separator for dataset table names, defaults to '___', i.e. 'database.dataset___table'."""
    table_engine_type: Optional[TTableEngineType] = "merge_tree"
    """The default table engine to use. Defaults to 'merge_tree'. Other implemented options are 'shared_merge_tree' and 'replicated_merge_tree'."""
    dataset_sentinel_table_name: str = "dlt_sentinel_table"
    """Special table to mark dataset as existing"""
    staging_use_https: bool = True
    """Connect to the staging buckets via https"""
    select_sequential_consistency: int = 1
    """Ensures read-after-write consistency on ClickHouse Cloud and clusters"""

    __config_gen_annotations__: ClassVar[List[str]] = [
        "dataset_table_separator",
        "dataset_sentinel_table_name",
        "table_engine_type",
    ]

    def fingerprint(self) -> str:
        """Returns a fingerprint of the host part of a connection string."""
        if self.credentials and self.credentials.host:
            return digest128(self.credentials.host)
        return ""



from typing import Any, Type
from dlt.common.configuration.exceptions import ConfigurationException


class SpecException(ConfigurationException):
    pass


class OAuth2ScopesRequired(SpecException):
    def __init__(self, spec: type) -> None:
        self.spec = spec
        super().__init__(
            "Scopes are required to retrieve refresh_token. Use `openid` scope for a token without"
            " any permissions to resources."
        )


class NativeValueError(SpecException, ValueError):
    def __init__(self, spec: Type[Any], native_value: str, msg: str) -> None:
        self.spec = spec
        self.native_value = native_value
        super().__init__(msg)


class InvalidConnectionString(NativeValueError):
    def __init__(self, spec: Type[Any], native_value: str, driver: str):
        driver = driver or "driver"
        msg = (
            f"The expected representation for `{spec.__name__}` is a standard database connection"
            f" string with the following format: `{driver}://username:password@host:port/database`"
        )
        super().__init__(spec, native_value, msg)


class InvalidGoogleNativeCredentialsType(NativeValueError):
    def __init__(self, spec: Type[Any], native_value: Any):
        msg = (
            f"Credentials `{spec.__name__}` accept a string with serialized credentials json file"
            " or an instance of `Credentials` object from Google.* namespace. The value passed is"
            f" of type `{type(native_value)}`"
        )
        super().__init__(spec, native_value, msg)


class InvalidGoogleServicesJson(NativeValueError):
    def __init__(self, spec: Type[Any], native_value: Any):
        msg = (
            f"The expected representation for `{spec.__name__}` is a string with serialized service"
            " account credentials, where at least `project_id`, `private_key` and `client_email`"
            " keys are present"
        )
        super().__init__(spec, native_value, msg)


class InvalidGoogleOauth2Json(NativeValueError):
    def __init__(self, spec: Type[Any], native_value: Any):
        msg = (
            f"The expected representation for `{spec.__name__}` is a string with serialized oauth2"
            " user info and may be wrapped in 'install'/'web' node - depending of oauth2 app type."
        )
        super().__init__(spec, native_value, msg)


class InvalidBoto3Session(NativeValueError):
    def __init__(self, spec: Type[Any], native_value: Any):
        msg = (
            f"The expected representation for `{spec.__name__}` is and instance of boto3.Session"
            " containing credentials"
        )
        super().__init__(spec, native_value, msg)


class ObjectStoreRsCredentialsException(ConfigurationException):
    pass


class UnsupportedAuthenticationMethodException(ConfigurationException):
    pass



from typing import ClassVar, Optional, Sequence
from dlt.common.configuration.specs import known_sections
from dlt.common.configuration.specs.base_configuration import BaseConfiguration, configspec
from dlt.common.configuration.specs.runtime_configuration import RuntimeConfiguration
from dlt.common.typing import TSecretStrValue


@configspec
class WorkspaceSettings(BaseConfiguration):
    name: Optional[str] = None
    # TODO: implement default profile switch. it requires reading the configuration, discovering
    # the profile and then recreating the workspace context. since this functionality is not
    # immediately needed it will be skipped for now
    # default_profile: Optional[str] = None
    working_dir: Optional[str] = None
    """Pipeline working dirs, other writable folders, local destination files (by default). Relative to workspace root"""
    local_dir: Optional[str] = None
    """Destination local files, by default it is within data_dir/local. Relative to workspace root"""


@configspec
class WorkspaceRuntimeConfiguration(RuntimeConfiguration):
    """Extends runtime configuration with dlthub runtime"""

    workspace_id: Optional[str] = None
    """Id of the remote workspace that local one should be connected to"""
    auth_token: Optional[TSecretStrValue] = None
    """JWT token for Runtime API"""
    auth_base_url: Optional[str] = "https://dlthub.app/api/auth"
    """Base URL for the dltHub Runtime authentication API"""
    api_base_url: Optional[str] = "https://dlthub.app/api/api"
    """Base URL for the dltHub Runtime API"""
    ai_context_api_url: Optional[str] = "https://scaffold.apps.dlthub.com"
    """Base URL for the AI context documentation API"""
    invite_code: Optional[str] = None
    """Invite code for dltHub Runtime"""

    __section__: ClassVar[str] = "runtime"


@configspec
class WorkspaceConfiguration(BaseConfiguration):
    settings: WorkspaceSettings = None
    runtime: WorkspaceRuntimeConfiguration = None
    # NOTE: is resolved separately but in the same layout
    # dashboard: DashboardConfiguration
    # TODO: launch workspace mcp using mcp configuration
    # mcp_config: McpConfiguration

    __recommended_sections__: ClassVar[Sequence[str]] = (known_sections.WORKSPACE,)



import os
import dataclasses
from typing import TYPE_CHECKING, Any, Dict, Optional, Final, Literal, ClassVar, List

from dlt.common import logger
from dlt.common.configuration import configspec
from dlt.common.configuration.specs.base_configuration import (
    BaseConfiguration,
    CredentialsConfiguration,
    NotResolved,
)
from dlt.common.destination.client import DestinationClientDwhConfiguration
from dlt.common.pendulum import timedelta
from dlt.common.storages.configuration import FilesystemConfiguration, WithLocalFiles
from dlt.common.typing import TSecretStrValue, Annotated
from dlt.common.utils import digest128

from dlt.destinations.impl.lancedb.warnings import uri_on_credentials_deprecated

if TYPE_CHECKING:
    from lancedb import DBConnection


@configspec
class LanceDBCredentials(CredentialsConfiguration):
    uri: Optional[str] = None
    api_key: Optional[TSecretStrValue] = None
    """API key for the remote connections (LanceDB cloud)."""
    region: Optional[str] = "us-east-1"
    """Region on LanceDB cloud"""
    host_override: Optional[str] = None
    read_consistency_interval: Optional[float] = None
    """Read consistency in seconds, None disables consistency"""
    client_config: Optional[Dict[str, Any]] = None
    storage_options: Optional[Dict[str, Any]] = None
    """Rust object store credentials"""

    embedding_model_provider_api_key: Optional[str] = None
    """API key for the embedding model provider."""

    _conn: Annotated[Optional["DBConnection"], NotResolved()] = None

    # TODO: we should refactor this similarly to duckdb or sqlalchemy: one connection with refcounting
    #   for all threads. we create separate clients to load each table which is not optimal
    def get_conn(self) -> "DBConnection":
        if self._conn:
            return self._conn

        import lancedb

        return lancedb.connect(
            uri=self.uri,
            api_key=self.api_key,
            region=self.region,
            host_override=self.host_override,
            read_consistency_interval=(
                None
                if self.read_consistency_interval is None
                else timedelta(seconds=self.read_consistency_interval)
            ),
            client_config=self.client_config,
            storage_options=self.storage_options,
        )

    def parse_native_representation(self, native_value: Any) -> None:
        try:
            # check if database was passed as explicit connection
            import lancedb

            if isinstance(native_value, lancedb.DBConnection):
                self._conn = native_value
                self.uri = ":external:"
                return
        except ImportError:
            pass

    __config_gen_annotations__: ClassVar[List[str]] = [
        "api_key",
        "embedding_model_provider_api_key",
    ]


@configspec
class LanceDBClientOptions(BaseConfiguration):
    max_retries: Optional[int] = 3
    """`EmbeddingFunction` class wraps the calls for source and query embedding
    generation inside a rate limit handler that retries the requests with exponential
    backoff after successive failures.

    You can tune it by setting it to a different number, or disable it by setting it to 0."""

    __config_gen_annotations__: ClassVar[List[str]] = [
        "max_retries",
    ]


TEmbeddingProvider = Literal[
    "gemini-text",
    "bedrock-text",
    "cohere",
    "gte-text",
    "imagebind",
    "instructor",
    "open-clip",
    "openai",
    "sentence-transformers",
    "huggingface",
    "colbert",
    "ollama",
]


@configspec
class LanceDBClientConfiguration(WithLocalFiles, DestinationClientDwhConfiguration):
    destination_type: Final[str] = dataclasses.field(  # type: ignore
        default="lancedb", init=False, repr=False, compare=False
    )
    lance_uri: Optional[str] = None
    """LanceDB database URI. Defaults to local, on-disk instance.

    The available schemas are:

    - `/path/to/database` - local database.
    - `db://host:port` - remote database (LanceDB cloud).
    """
    credentials: LanceDBCredentials = None
    dataset_separator: str = "___"
    """Character for the dataset separator."""
    dataset_name: Final[Optional[str]] = dataclasses.field(  # type: ignore
        default=None, init=False, repr=False, compare=False
    )

    options: Optional[LanceDBClientOptions] = None
    """LanceDB client options."""

    embedding_model_provider: TEmbeddingProvider = "cohere"
    """Embedding provider used for generating embeddings. Default is "cohere". You can find the full list of
    providers at https://github.com/lancedb/lancedb/tree/main/python/python/lancedb/embeddings as well as
    https://lancedb.github.io/lancedb/embeddings/default_embedding_functions/."""
    embedding_model_provider_host: Optional[str] = None
    """Full host URL with protocol and port (e.g. 'http://localhost:11434'). Uses LanceDB's default if not specified, assuming the provider accepts this parameter."""
    embedding_model: str = "embed-english-v3.0"
    """The model used by the embedding provider for generating embeddings.
    Check with the embedding provider which options are available.
    Reference https://lancedb.github.io/lancedb/embeddings/default_embedding_functions/."""
    embedding_model_dimensions: Optional[int] = None
    """The dimensions of the embeddings generated. In most cases it will be automatically inferred, by LanceDB,
    but it is configurable in rare cases.

    Make sure it corresponds with the associated embedding model's dimensionality."""
    vector_field_name: str = "vector"
    """Name of the special field to store the vector embeddings."""
    sentinel_table_name: str = "dltSentinelTable"
    """Name of the sentinel table that encapsulates datasets. Since LanceDB has no
    concept of schemas, this table serves as a proxy to group related dlt tables together."""

    __config_gen_annotations__: ClassVar[List[str]] = [
        "lance_uri",
        "embedding_model",
        "embedding_model_provider",
    ]

    def on_resolved(self) -> None:
        try:
            from lancedb import DBConnection

            if isinstance(self.credentials, DBConnection):
                if self.lance_uri:
                    logger.warning(
                        "Lance db client was passed directly so `lance_uri` with database location"
                        " is ignored"
                    )
                self.lance_uri = ":external:"
                return
        except ImportError:
            # native client not possible if dependency not found
            pass

        if self.credentials.uri and not self.lance_uri:
            uri_on_credentials_deprecated()
            self.lance_uri = self.credentials.uri

        # use local path if uri not provided or relative path
        if (
            not self.lance_uri
            or FilesystemConfiguration.is_local_path(self.lance_uri)
            and not os.path.isabs(self.lance_uri)
        ):
            self.lance_uri = self.make_location(self.lance_uri, "%s.lancedb")
        # TODO: move uri back to credentials to make it more like other connections
        self.credentials.uri = self.lance_uri

    def fingerprint(self) -> str:
        """Returns a fingerprint of a connection string."""

        if self.lance_uri:
            return digest128(self.lance_uri)
        return ""



from typing import ClassVar, Type

from duckdb import DuckDBPyConnection

from dlt.common import logger
from dlt.common.configuration.specs.connection_string_credentials import ConnectionStringCredentials
from dlt.common.destination.capabilities import DestinationCapabilitiesContext
from dlt.common.storages.configuration import FileSystemCredentials
from dlt.common.storages.fsspec_filesystem import fsspec_from_config
from dlt.destinations.impl.duckdb.configuration import DuckDbCredentials
from dlt.destinations.impl.duckdb.sql_client import DuckDBDBApiCursorImpl, DuckDbSqlClient
from dlt.destinations.impl.ducklake.configuration import DuckLakeCredentials
from dlt.destinations.sql_client import raise_open_connection_error


class DuckLakeDBApiCursorImpl(DuckDBDBApiCursorImpl):
    vector_size: ClassVar[int] = 700  # vector size for ducklake


class DuckLakeSqlClient(DuckDbSqlClient):
    cursor_impl: ClassVar[Type[DuckDBDBApiCursorImpl]] = DuckLakeDBApiCursorImpl

    def __init__(
        self,
        dataset_name: str,
        staging_dataset_name: str,
        credentials: DuckLakeCredentials,
        capabilities: DestinationCapabilitiesContext,
        override_data_path: bool = False,
    ) -> None:
        super().__init__(dataset_name, staging_dataset_name, credentials, capabilities)
        self.credentials: DuckLakeCredentials = credentials
        self._attach_statement: str = None
        self.override_data_path = override_data_path

    def create_dataset(self) -> None:
        if self.has_dataset():
            import duckdb

            raise self._make_database_exception(
                duckdb.CatalogException(
                    f'Catalog Error: Schema with name "{self.fully_qualified_dataset_name()}"'
                    " already exists!"
                )
            )
        super().create_dataset()

    @raise_open_connection_error
    def open_connection(self) -> DuckDBPyConnection:
        """Ensure the `ducklake` extension is loaded. This needs to be done on every connection"""
        super().open_connection()
        # setup secrets and attach ducklake for each opened connections. connection pool
        # creates a separate connection for each sql_client
        try:
            if not self.credentials.storage.is_local_filesystem:
                self.create_secret(
                    self.credentials.storage.bucket_url, self.credentials.storage.credentials
                )
            # NOTE: database must be detached otherwise it is left in inconsistent state
            # TODO: perhaps move attach/detach to connection pool
            self._conn.execute(self.attach_statement)
            self._conn.execute(
                f"USE {self.capabilities.escape_identifier(self.credentials.ducklake_name)};"
            )
            # search path can only by set after database is attached
            try:
                self._conn.execute(f"SET search_path = '{self.fully_qualified_dataset_name()}'")
            except Exception:
                pass
        except Exception:
            self.close_connection()
            raise
        return self._conn

    def close_connection(self) -> None:
        if self._conn:
            # metadata database MUST be detached properly before connection is closed, we observed
            # corrupted duckdb catalogs if not done properly
            try:
                # rollback any pending transaction
                self._conn.execute("ROLLBACK;")
            except Exception:
                pass
            # make sure catalog is attached
            current_db = self._conn.sql("SELECT current_database();").fetchone()[0]
            if current_db == self.credentials.ducklake_name:
                # `memory` is a name of initial memory database opened
                self._conn.execute(f"USE memory;DETACH {self.credentials.ducklake_name}")
        return super().close_connection()

    def create_secret(
        self,
        scope: str,
        credentials: FileSystemCredentials,
        secret_name: str = None,
        persist_secrets: bool = False,
    ) -> None:
        protocol = self.credentials.storage.protocol
        if protocol in ["az", "abfss"]:
            logger.warning(
                "abfss is not supported by DuckLake. "
                "Falling back to fsspec which degrades scanning performance."
            )
            self._register_filesystem(fsspec_from_config(self.credentials.storage)[0], "abfss")
        elif not super().create_secret(
            scope,
            credentials,
            secret_name=secret_name,
            persist_secrets=persist_secrets,
        ):
            if protocol in ["gs", "gcs"]:
                logger.warning(
                    "For gs/gcs access via duckdb please use the gs/gcs s3 compatibility"
                    "layer if possible. "
                    "Falling back to fsspec which degrades scanning performance."
                )
                self._register_filesystem(fsspec_from_config(self.credentials.storage)[0], "gcs")
            else:
                raise ValueError(f"Cannot create secret or register filesystem for `{protocol=}`")

    @staticmethod
    def build_attach_statement(
        *,
        ducklake_name: str,
        catalog: ConnectionStringCredentials,
        storage_url: str,
        override_data_path: bool = False,
    ) -> str:
        attach_params = ""
        if isinstance(catalog, DuckDbCredentials):
            attach_statement = f"ATTACH IF NOT EXISTS 'ducklake:{catalog._conn_str()}'"
        elif catalog.drivername in ("postgres", "postgresql", "mysql"):
            # `drivername="postgresql"` is supported by dlt / sqlalchemy, but it doesn't exist in duckdb.
            if catalog.drivername == "postgresql":
                catalog.drivername = "postgres"

            db_url = catalog.to_url().render_as_string(hide_password=False)
            attach_statement = f"ATTACH IF NOT EXISTS 'ducklake:{catalog.drivername}:{db_url}'"
            attach_params = f", METADATA_SCHEMA '{ducklake_name}'"
        elif catalog.drivername == "md":
            logger.warning(
                "Motherduck requires token present in the environment and will most probably crash."
            )
            attach_statement = f"ATTACH IF NOT EXISTS 'ducklake:md:{catalog.database}'"
            attach_params = f", METADATA_SCHEMA '{ducklake_name}'"
        elif catalog.drivername in ("sqlite", "duckdb"):
            # attach sqllite with multi-process access
            attach_statement = f"ATTACH IF NOT EXISTS 'ducklake:{catalog.database}'"
            attach_params = ", META_TYPE 'sqlite', META_JOURNAL_MODE 'WAL', META_BUSY_TIMEOUT 1000"
        else:
            raise NotImplementedError(str(catalog))
        attach_statement += f" AS {ducklake_name}"
        override_param = ", OVERRIDE_DATA_PATH true" if override_data_path else ""
        attach_statement += f" (DATA_PATH '{storage_url}'{attach_params}{override_param})"
        return attach_statement

    @property
    def attach_statement(self) -> str:
        # return value when set explicitly
        if self._attach_statement:
            return self._attach_statement
        else:
            return self.build_attach_statement(
                ducklake_name=self.credentials.ducklake_name,
                catalog=self.credentials.catalog,
                storage_url=self.credentials.storage_url,
                override_data_path=self.override_data_path,
            )



from typing import Any, Type, TYPE_CHECKING, Optional

from dlt.common import logger
from dlt.common.destination.typing import PreparedTableSchema
from dlt.common.exceptions import TerminalValueError
from dlt.common.normalizers.naming import NamingConvention
from dlt.common.configuration.specs import GcpServiceAccountCredentials
from dlt.common.arithmetics import DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE
from dlt.common.data_writers.escape import (
    escape_bigquery_identifier,
    escape_bigquery_literal,
    format_bigquery_datetime_literal,
)
from dlt.common.destination import Destination, DestinationCapabilitiesContext
from dlt.common.schema.typing import TColumnSchema, TColumnType
from dlt.common.typing import TLoaderFileFormat

from dlt.destinations.type_mapping import TypeMapperImpl
from dlt.destinations.impl.bigquery.bigquery_adapter import should_autodetect_schema
from dlt.destinations.impl.bigquery.configuration import BigQueryClientConfiguration
from dlt.destinations.utils import parse_db_data_type_str_with_precision


if TYPE_CHECKING:
    from dlt.destinations.impl.bigquery.bigquery import BigQueryClient


class BigQueryTypeMapper(TypeMapperImpl):
    sct_to_unbound_dbt = {
        "json": "JSON",
        "text": "STRING",
        "double": "FLOAT64",
        "bool": "BOOL",
        "date": "DATE",
        "timestamp": "TIMESTAMP",
        "bigint": "INT64",
        "binary": "BYTES",
        "wei": "BIGNUMERIC",  # non-parametrized should hold wei values
        "time": "TIME",
    }

    sct_to_dbt = {
        "text": "STRING(%i)",
        "binary": "BYTES(%i)",
    }

    dbt_to_sct = {
        "STRING": "text",
        "FLOAT64": "double",
        "BOOL": "bool",
        "DATE": "date",
        "TIMESTAMP": "timestamp",
        "DATETIME": "timestamp",
        "INT64": "bigint",
        "BYTES": "binary",
        "NUMERIC": "decimal",
        "BIGNUMERIC": "decimal",
        "JSON": "json",
        "TIME": "time",
    }

    def ensure_supported_type(
        self,
        column: TColumnSchema,
        table: PreparedTableSchema,
        loader_file_format: TLoaderFileFormat,
    ) -> None:
        # if table contains json types, we cannot load with parquet
        if (
            loader_file_format == "parquet"
            and column["data_type"] == "json"
            and not should_autodetect_schema(table)
        ):
            raise TerminalValueError(
                "Enable `autodetect_schema` in config or via BigQuery adapter", column["data_type"]
            )

    def to_db_datetime_type(
        self,
        column: TColumnSchema,
        table: PreparedTableSchema = None,
    ) -> str:
        column_name = column["name"]
        table_name = table["name"]
        precision = column.get("precision")

        if precision and precision != self.capabilities.timestamp_precision:
            logger.warn(
                f"BigQuery does not support custom precision for column '{column_name}' in"
                f" table '{table_name}'. Will use default precision."
            )

        # TIMESTAMP is always timezone-aware in BigQuery
        # DATETIME is always timezone-naive in BigQuery
        # NOTE: we disable DATETIME because it does not work with parquet
        # timezone = column.get("timezone", True)
        return "TIMESTAMP"

    def to_db_decimal_type(self, column: TColumnSchema) -> str:
        # Use BigQuery's BIGNUMERIC for large precision decimals
        precision, scale = self.decimal_precision(column.get("precision"), column.get("scale"))

        # BigQuery NUMERIC constraints:
        # - When scale = 0: precision must be 1-29
        # - When scale > 0: precision can be up to 38
        # - scale must be <= 9
        if scale == 0 and precision > 29:
            return "BIGNUMERIC(%i,%i)" % (precision, scale)
        elif precision > 38 or scale > 9:
            return "BIGNUMERIC(%i,%i)" % (precision, scale)
        return "NUMERIC(%i,%i)" % (precision, scale)

    # noinspection PyTypeChecker,PydanticTypeChecker
    def from_destination_type(
        self, db_type: str, precision: Optional[int], scale: Optional[int]
    ) -> TColumnType:
        # precision is present in the type name
        if db_type == "BIGNUMERIC":
            return dict(data_type="wei")
        if db_type == "DATETIME":
            return {"data_type": "timestamp", "timezone": False}
        return super().from_destination_type(*parse_db_data_type_str_with_precision(db_type))


# noinspection PyPep8Naming
class bigquery(Destination[BigQueryClientConfiguration, "BigQueryClient"]):
    spec = BigQueryClientConfiguration

    def _raw_capabilities(self) -> DestinationCapabilitiesContext:
        caps = DestinationCapabilitiesContext()
        caps.preferred_loader_file_format = "jsonl"
        caps.supported_loader_file_formats = ["jsonl", "parquet", "model"]
        caps.preferred_staging_file_format = "parquet"
        caps.supported_staging_file_formats = ["parquet", "jsonl"]
        caps.type_mapper = BigQueryTypeMapper
        # BigQuery is by default case sensitive but that cannot be turned off for a dataset
        # https://cloud.google.com/bigquery/docs/reference/standard-sql/lexical#case_sensitivity
        caps.escape_identifier = escape_bigquery_identifier
        caps.escape_literal = escape_bigquery_literal
        caps.has_case_sensitive_identifiers = True
        caps.casefold_identifier = str
        # BQ limit is 4GB but leave a large headroom since buffered writer does not preemptively check size
        caps.recommended_file_size = int(1024 * 1024 * 1024)
        caps.format_datetime_literal = format_bigquery_datetime_literal
        caps.decimal_precision = (DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE)
        caps.wei_precision = (76, 38)
        caps.max_identifier_length = 1024
        caps.max_column_identifier_length = 300
        caps.max_query_length = 1024 * 1024
        caps.is_max_query_length_in_bytes = False
        caps.max_text_data_type_length = 10 * 1024 * 1024
        caps.is_max_text_data_type_length_in_bytes = True
        caps.supports_ddl_transactions = False
        caps.supports_clone_table = True
        caps.supports_naive_datetime = False
        caps.schema_supports_numeric_precision = False  # no precision information in BigQuery
        caps.supported_merge_strategies = ["delete-insert", "upsert", "scd2"]
        caps.supported_replace_strategies = [
            "truncate-and-insert",
            "insert-from-staging",
            "staging-optimized",
        ]
        caps.sqlglot_dialect = "bigquery"

        return caps

    @property
    def client_class(self) -> Type["BigQueryClient"]:
        from dlt.destinations.impl.bigquery.bigquery import BigQueryClient

        return BigQueryClient

    def __init__(
        self,
        credentials: GcpServiceAccountCredentials = None,
        location: str = None,
        has_case_sensitive_identifiers: bool = None,
        destination_name: str = None,
        environment: str = None,
        **kwargs: Any,
    ) -> None:
        """Configure the BigQuery destination to use in a pipeline.

        All arguments provided here supersede other configuration sources such as environment variables and dlt config files.

        Args:
            credentials (GcpServiceAccountCredentials, optional): Credentials to connect to the BigQuery database. Can be an instance of `GcpServiceAccountCredentials` or
                a dict or string with service accounts credentials as used in the Google Cloud
            location (str, optional): A location where the datasets will be created, eg. "EU". The default is "US"
            has_case_sensitive_identifiers (bool, optional): Is the dataset case-sensitive, defaults to True
            destination_name (str, optional): Name of the destination, can be used in config section to differentiate between multiple of the same type
            environment (str, optional): Environment of the destination
            **kwargs (Any): Additional arguments passed to the destination config
        """
        super().__init__(
            credentials=credentials,
            location=location,
            has_case_sensitive_identifiers=has_case_sensitive_identifiers,
            destination_name=destination_name,
            environment=environment,
            **kwargs,
        )

    @classmethod
    def adjust_capabilities(
        cls,
        caps: DestinationCapabilitiesContext,
        config: BigQueryClientConfiguration,
        naming: Optional[NamingConvention],
    ) -> DestinationCapabilitiesContext:
        # modify the caps if case sensitive identifiers are requested
        if config.should_set_case_sensitivity_on_new_dataset:
            caps.has_case_sensitive_identifiers = config.has_case_sensitive_identifiers
        return super().adjust_capabilities(caps, config, naming)


bigquery.register()



"""Constants and type aliases for the dashboard."""

from typing import Dict, List, Literal

from typing_extensions import TypeAlias

from dlt.common.storages.load_package import TLoadPackageStatus


# Runner file names

EJECTED_APP_FILE_NAME = "dlt_dashboard.py"
STYLE_FILE_NAME = "dlt_dashboard_styles.css"

# Runner defaults

DEFAULT_DASHBOARD_PORT = 2718
"""Default port for the marimo dashboard server."""

MARIMO_COMMAND = "marimo"
"""The marimo CLI executable name."""

HTTP_READY_TIMEOUT_S = 15.0
"""Seconds to wait for the dashboard HTTP server to become ready."""

HTTP_READY_POLL_INTERVAL_S = 0.1
"""Seconds between HTTP readiness polls."""

HTTP_READY_WAIT_ON_OK_S = 0.1
"""Extra wait after the server responds OK (default for _wait_http_up)."""

DASHBOARD_STARTUP_TIMEOUT_S = 60.0
"""Seconds to wait for the dashboard process in start_dashboard()."""

DASHBOARD_STARTUP_WAIT_ON_OK_S = 1.0
"""Extra wait after ready in start_dashboard()."""

PROCESS_TERMINATE_TIMEOUT_S = 10
"""Seconds to wait for the dashboard process to terminate before killing it."""


# DashboardConfiguration defaults

DEFAULT_TABLE_LIST_FIELDS: List[str] = [
    "parent",
    "resource",
    "write_disposition",
    "description",
]
"""Default fields shown in table lists (name is always present)."""

DEFAULT_COLUMN_TYPE_HINTS: List[str] = [
    "data_type",
    "nullable",
    "precision",
    "scale",
    "timezone",
    "variant",
]
"""Default column type hints shown in the column list."""

DEFAULT_COLUMN_OTHER_HINTS: List[str] = [
    "primary_key",
    "merge_key",
    "unique",
]
"""Default other column hints shown in the column list."""

DEFAULT_DATETIME_FORMAT = "YYYY-MM-DD HH:mm:ss Z"
"""Default datetime display format."""


# Pipeline execution visualization

TPipelineRunStatus: TypeAlias = Literal["succeeded", "failed"]
TVisualPipelineStep: TypeAlias = Literal["extract", "normalize", "load"]

VISUAL_PIPELINE_STEPS: List[TVisualPipelineStep] = ["extract", "normalize", "load"]

PIPELINE_RUN_STEP_COLORS: Dict[TVisualPipelineStep, str] = {
    "extract": "var(--dlt-color-lime)",
    "normalize": "var(--dlt-color-aqua)",
    "load": "var(--dlt-color-pink)",
}


# Load package statuses

PENDING_LOAD_STATUSES: Dict[TLoadPackageStatus, str] = {
    "extracted": "pending to normalize",
    "normalized": "pending to load",
}

LOAD_PACKAGE_STATUS_COLORS: Dict[TLoadPackageStatus, str] = {
    "new": "grey",
    "extracted": "yellow",
    "normalized": "yellow",
    "loaded": "green",
    "aborted": "red",
}



from importlib.metadata import version as pkg_version, distribution as pkg_distribution
from typing import Optional
from urllib.request import url2pathname
from urllib.parse import urlparse
from packaging.requirements import Requirement

DLT_IMPORT_NAME = "dlt"
PKG_NAME = DLT_PKG_NAME = "dlt"
__version__ = pkg_version(DLT_PKG_NAME)
DLT_PKG_REQUIREMENT = f"{DLT_PKG_NAME}=={__version__}"


def get_installed_requirement_string(
    package: str = DLT_PKG_NAME, allow_earlier: bool = False
) -> str:
    """Gets the requirement string of currently installed dlt version"""
    dist = pkg_distribution(package)
    # PEP 610 https://packaging.python.org/en/latest/specifications/direct-url/#specification
    direct_url = dist.read_text("direct_url.json")
    if direct_url is not None:
        from dlt.common.json import json

        # `url` contain the location of the distribution
        url = urlparse(json.loads(direct_url)["url"])
        # we are interested only in file urls
        if url.scheme == "file":
            return url2pathname(url.path)

    ver_selector = "<=" if allow_earlier else "=="
    if package == DLT_PKG_NAME:
        package_requirement = f"{DLT_PKG_NAME}{ver_selector}{__version__}"
    else:
        package_requirement = f"{package}{ver_selector}{pkg_version(package)}"
    return package_requirement


def get_dependency_requirement(
    dependency_name: str, package: str = DLT_PKG_NAME
) -> Optional[Requirement]:
    """Find a specific dependency requirement from package metadata"""
    dist = pkg_distribution(package)

    if dist.requires is None:
        return None

    for req_str in dist.requires:
        req = Requirement(req_str)
        if req.name == dependency_name:
            return req
    return None



from __future__ import annotations

import functools
from abc import ABC, abstractmethod
from typing import Iterator, TYPE_CHECKING, Sequence, Tuple

from dlt.common import logger
from dlt.common.configuration.inject import with_config
from dlt.common.destination.capabilities import LoaderFileFormatSelector
from dlt.common.schema.typing import TTableSchema
from dlt.common.typing import TLoaderFileFormat
from dlt.common.utils import without_none
from dlt.destinations.job_client_impl import SqlJobClientBase

if TYPE_CHECKING:
    from adbc_driver_manager.dbapi import Connection

from dlt.common.destination.client import HasFollowupJobs, RunnableLoadJob


# TODO: driver presence detection, driver location detection to support (see postgres factory)
# dbc and pip install drivers, connection string conversion etc. should be extracted to ADBC
# lib helper (like we do with sqlalchemy or arrow)


@with_config
def has_adbc_driver(driver: str, disable_adbc_detection: bool = False) -> Tuple[bool, str]:
    """Figures out if given driver is available without actually connecting to destination.
    Allows to disable via `disable_adbc_detection` setting in dlt config
    """
    if disable_adbc_detection:
        return False, None
    try:
        import adbc_driver_manager as dm

        try:
            db = dm.AdbcDatabase(driver=driver, uri="")
            db.close()
            return True, None
        except dm.Error as pex:
            # NOT_FOUND returned when driver library can't be found
            if pex.status_code in (dm.AdbcStatusCode.NOT_FOUND, dm.AdbcStatusCode.NOT_IMPLEMENTED):
                return False, str(pex)
            return True, str(pex)
    except ImportError as import_ex:
        return False, str(import_ex)


class AdbcParquetCopyJob(RunnableLoadJob, HasFollowupJobs, ABC):
    def __init__(self, file_path: str) -> None:
        super().__init__(file_path)
        self._job_client: SqlJobClientBase = None
        # override default schema handling
        self._connect_catalog_name: str = None
        self._connect_schema_name: str = None

    @abstractmethod
    def _connect(self) -> Connection:
        pass

    def _set_catalog_and_schema(self) -> Tuple[str, str]:
        catalog_name = self._connect_catalog_name
        if catalog_name is None:
            catalog_name = self._job_client.sql_client.catalog_name(quote=False)
        elif catalog_name == "":
            # empty string disables catalog
            catalog_name = None

        schema_name = self._connect_schema_name
        if schema_name is None:
            schema_name = self._job_client.sql_client.escape_column_name(
                self._job_client.sql_client.dataset_name, quote=False, casefold=True
            )
        elif schema_name == "":
            # empty string disables schema
            schema_name = None

        return catalog_name, schema_name

    def run(self) -> None:
        from dlt.common.libs.pyarrow import pq_stream_with_new_columns
        from dlt.common.libs.pyarrow import pyarrow

        def _iter_batches(file_path: str) -> Iterator[pyarrow.RecordBatch]:
            for table in pq_stream_with_new_columns(file_path, ()):
                yield from table.to_batches()

        with self._connect() as conn, conn.cursor() as cur:
            import time

            catalog_name, schema_name = self._set_catalog_and_schema()
            kwargs = dict(catalog_name=catalog_name, db_schema_name=schema_name)

            t_ = time.monotonic()
            rows = cur.adbc_ingest(
                self.load_table_name,
                _iter_batches(self._file_path),
                mode="append",
                **without_none(kwargs),  # type: ignore[arg-type,unused-ignore]
            )
            conn.commit()
            logger.info(
                f"{rows} rows copied from {self._file_name} to"
                f" {self.load_table_name}.{schema_name} in {time.monotonic()-t_} s"
            )


def _loader_file_format_selector(
    driver: str,
    docs_url: str,
    prefer_parquet: bool,
    preferred_loader_file_format: TLoaderFileFormat,
    supported_loader_file_formats: Sequence[TLoaderFileFormat],
    /,
    *,
    table_schema: TTableSchema,
) -> Tuple[TLoaderFileFormat, Sequence[TLoaderFileFormat]]:
    found, err_str = has_adbc_driver(driver)
    if not found:
        supported_loader_file_formats = list(supported_loader_file_formats)
        supported_loader_file_formats.remove("parquet")

        if table_schema.get("file_format") == "parquet":
            logger.warning(
                f"parquet file format was requested for table {table_schema['name']} but ADBC"
                f" driver for {driver} was not installed:\n {err_str}\n"
                " Read more: "
                + docs_url
            )
    else:
        if prefer_parquet:
            # parquet is preferred format if driver is enabled
            preferred_loader_file_format = "parquet"

    return (preferred_loader_file_format, supported_loader_file_formats)


def make_adbc_parquet_file_format_selector(
    driver: str,
    docs_url: str,
    prefer_parquet: bool,
) -> LoaderFileFormatSelector:
    """Factory for file format selector that removes parquet from the list if `driver` not installed"""

    return functools.partial(_loader_file_format_selector, driver, docs_url, prefer_parquet)  # type: ignore[return-value]



from typing import (
    Optional,
    Any,
    Sequence,
    Tuple,
    List,
    Dict,
    Iterable,
    Union,
    cast,
    TYPE_CHECKING,
)

from dlt.common.schema.exceptions import TableNotFound

if TYPE_CHECKING:
    from mypy_boto3_lakeformation import LakeFormationClient
    from mypy_boto3_lakeformation.type_defs import (
        AddLFTagsToResourceResponseTypeDef,
        GetResourceLFTagsResponseTypeDef,
        RemoveLFTagsFromResourceResponseTypeDef,
        ResourceTypeDef,
    )
else:
    LakeFormationClient = Any
    AddLFTagsToResourceResponseTypeDef = Any
    GetResourceLFTagsResponseTypeDef = Any
    RemoveLFTagsFromResourceResponseTypeDef = Any
    ResourceTypeDef = Any

from threading import Lock

from dlt.common import logger
from dlt.common.utils import uniq_id
from dlt.common.schema import TColumnSchema, Schema
from dlt.common.schema.typing import (
    TColumnType,
    TSchemaTables,
    TSortOrder,
    TTableSchema,
)
from dlt.common.destination import DestinationCapabilitiesContext, PreparedTableSchema
from dlt.common.destination.utils import resolve_merge_strategy
from dlt.common.destination.client import FollowupJobRequest, SupportsStagingDestination, LoadJob
from dlt.destinations.sql_jobs import (
    SqlStagingCopyFollowupJob,
    SqlStagingReplaceFollowupJob,
    SqlMergeFollowupJob,
)

from dlt.destinations import path_utils
from dlt.destinations.sql_client import SqlClientBase
from dlt.destinations.utils import get_deterministic_temp_table_name
from dlt.destinations.job_client_impl import SqlJobClientWithStagingDataset
from dlt.destinations.job_impl import FinalizedLoadJobWithFollowupJobs, FinalizedLoadJob
from dlt.destinations.impl.athena.configuration import (
    DEFAULT_TABLE_LOCATION_LAYOUT,
    AthenaClientConfiguration,
)
from dlt.destinations.impl.athena.athena_adapter import PARTITION_HINT
from dlt.destinations.impl.athena.configuration import LakeformationConfig
from dlt.destinations.impl.athena.sql_client import AthenaSQLClient
from dlt.destinations.impl.athena.utils import is_s3_tables_catalog


boto3_client_lock = Lock()


class AthenaMergeJob(SqlMergeFollowupJob):
    @classmethod
    def _new_temp_table_name(cls, table_name: str, op: str, sql_client: SqlClientBase[Any]) -> str:
        # reproducible name so we know which table to drop
        with sql_client.with_staging_dataset():
            return sql_client.make_qualified_table_name(
                cls._shorten_table_name(
                    get_deterministic_temp_table_name(table_name, op), sql_client
                )
            )

    @classmethod
    def _to_temp_table(cls, select_sql: str, temp_table_name: str, unique_column: str) -> str:
        # regular table because Athena does not support temporary tables
        return f"CREATE TABLE {temp_table_name} AS {select_sql}"

    @classmethod
    def gen_insert_temp_table_sql(
        cls,
        table_name: str,
        staging_root_table_name: str,
        sql_client: SqlClientBase[Any],
        primary_keys: Sequence[str],
        unique_column: str,
        dedup_sort: Tuple[str, TSortOrder] = None,
        condition: str = None,
        condition_columns: Sequence[str] = None,
        skip_dedup: bool = False,
    ) -> Tuple[List[str], str]:
        sql, temp_table_name = super().gen_insert_temp_table_sql(
            table_name,
            staging_root_table_name,
            sql_client,
            primary_keys,
            unique_column,
            dedup_sort,
            condition,
            condition_columns,
            skip_dedup,
        )
        # DROP needs backtick as escape identifier
        sql.insert(0, f"""DROP TABLE IF EXISTS {temp_table_name.replace('"', '`')}""")
        return sql, temp_table_name

    @classmethod
    def gen_delete_temp_table_sql(
        cls,
        table_name: str,
        unique_column: str,
        key_table_clauses: Sequence[str],
        sql_client: SqlClientBase[Any],
    ) -> Tuple[List[str], str]:
        sql, temp_table_name = super().gen_delete_temp_table_sql(
            table_name, unique_column, key_table_clauses, sql_client
        )
        # DROP needs backtick as escape identifier
        sql.insert(0, f"""DROP TABLE IF EXISTS {temp_table_name.replace('"', '`')}""")
        return sql, temp_table_name

    @classmethod
    def gen_concat_sql(cls, columns: Sequence[str]) -> str:
        # Athena requires explicit casting
        columns = [f"CAST({c} AS VARCHAR)" for c in columns]
        return f"CONCAT({', '.join(columns)})"

    @classmethod
    def requires_temp_table_for_delete(cls) -> bool:
        return True


class LfTagsManager:
    """Class used to manage lakeformation tags on glue/athena database.

    Heavily inspired by dbt-athena implementation:
    https://github.com/dbt-labs/dbt-athena/blob/main/dbt-athena/src/dbt/adapters/athena/lakeformation.py
    """

    def __init__(
        self,
        lf_client: LakeFormationClient,
        lf_config: LakeformationConfig,
        catalog: str,
        dataset: str,
        table: Optional[str] = None,
    ):
        self.lf_client = lf_client
        self.catalog = catalog
        self.database = dataset
        self.table = table
        self.enabled = lf_config.enabled
        self.lf_tags = lf_config.tags

    def process_lf_tags_database(self) -> None:
        """Add or remove lf tags from athena database."""
        database_resource = self._database_resource
        if self.enabled and self.lf_tags:
            add_response = self.lf_client.add_lf_tags_to_resource(
                Resource=database_resource,
                LFTags=[{"TagKey": k, "TagValues": [v]} for k, v in self.lf_tags.items()],
            )
            self._parse_and_log_lf_response(add_response, None, self.lf_tags)
        elif not self.enabled:
            get_response: GetResourceLFTagsResponseTypeDef = self.lf_client.get_resource_lf_tags(
                Resource=database_resource
            )
            if "LFTagOnDatabase" not in get_response:
                return
            tags = get_response["LFTagOnDatabase"]
            self.lf_client.remove_lf_tags_from_resource(Resource=database_resource, LFTags=tags)

        else:
            logger.debug("Lakeformation is enabled but no tags are set")

    @property
    def _database_resource(self) -> ResourceTypeDef:
        return self._get_database_resource(self.database, self.catalog)

    @staticmethod
    def _get_database_resource(database_name: str, catalog_id: str) -> ResourceTypeDef:
        database_dict = {"Name": database_name}
        # NOTE: LakeFormation's default catalog is the AWS account ID, while Athena's default
        # catalog is "awsdatacatalog"; hence, setting CatalogId to "awsdatacatalog" will error.
        # Luckily, we only need to set CatalogId if not using LakeFormation's default catalog.
        if is_s3_tables_catalog(catalog_id):
            database_dict["CatalogId"] = catalog_id
        database_resource = {"Database": database_dict}
        return cast(ResourceTypeDef, database_resource)

    def _parse_and_log_lf_response(
        self,
        response: Union[
            AddLFTagsToResourceResponseTypeDef, RemoveLFTagsFromResourceResponseTypeDef
        ],
        columns: Optional[List[str]] = None,
        lf_tags: Optional[Dict[str, str]] = None,
        verb: str = "add",
    ) -> None:
        table_appendix = f".{self.table}" if self.table else ""
        columns_appendix = f" for columns {columns}" if columns else ""
        resource_msg = self.database + table_appendix + columns_appendix
        if failures := response.get("Failures", []):
            base_msg = f"Failed to {verb} LF tags: {lf_tags} to " + resource_msg
            for failure in failures:
                tag = failure.get("LFTag", {}).get("TagKey")
                error = failure.get("Error", {}).get("ErrorMessage")
                logger.error(f"Failed to {verb} {tag} for " + resource_msg + f" - {error}")
            raise RuntimeError(base_msg)
        logger.debug(f"Success: {verb} LF tags {lf_tags} to " + resource_msg)


class AthenaClient(SqlJobClientWithStagingDataset, SupportsStagingDestination):
    def __init__(
        self,
        schema: Schema,
        config: AthenaClientConfiguration,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        # verify if staging layout is valid for Athena
        # this will raise if the table prefix is not properly defined
        # we actually make sure that {table_name} is first, no {schema_name} is allowed
        if config.staging_config:
            self.table_prefix_layout = path_utils.get_table_prefix_layout(
                config.staging_config.layout,
                supported_prefix_placeholders=[],
                table_needs_own_folder=True,
            )

        dataset_name, staging_dataset_name = SqlJobClientWithStagingDataset.create_dataset_names(
            schema, config
        )

        sql_client = AthenaSQLClient(
            dataset_name,
            staging_dataset_name,
            config,
            capabilities,
        )
        super().__init__(schema, config, sql_client)
        self.sql_client: AthenaSQLClient = sql_client  # type: ignore
        self.config: AthenaClientConfiguration = config
        self.type_mapper = self.capabilities.get_type_mapper()
        self.allow_merge_to_append_fallback = True

    def initialize_storage(self, truncate_tables: Iterable[str] = None) -> None:
        # truncate only iceberg tables via SQL (DELETE FROM ... WHERE 1=1)
        # non-iceberg (HIVE external) tables have data on staging filesystem
        if truncate_tables:
            truncate_tables = [
                t
                for t in truncate_tables
                if self._is_iceberg_table(self.prepare_load_table(t), self.in_staging_dataset_mode)
            ]
        super().initialize_storage(truncate_tables or None)

    def prepare_load_table(self, table_name: str) -> PreparedTableSchema:
        load_table = super().prepare_load_table(table_name)
        if self._is_s3_table():
            load_table["table_format"] = "iceberg"
        if load_table["write_disposition"] == "merge":
            merge_strategy = resolve_merge_strategy(
                self.schema.tables, load_table, self.capabilities
            )
            if merge_strategy is None:
                logger.warning(f"Falling back to `append` on `{table_name}`.")
                load_table["write_disposition"] = "append"
        return load_table

    def _from_db_type(
        self, hive_t: str, precision: Optional[int], scale: Optional[int]
    ) -> TColumnType:
        return self.type_mapper.from_destination_type(hive_t, precision, scale)

    def _get_column_def_sql(self, c: TColumnSchema, table: PreparedTableSchema = None) -> str:
        return (
            f"{self.sql_client.escape_ddl_identifier(c['name'])} {self.type_mapper.to_destination_type(c, table)}"
        )

    def _iceberg_partition_clause(self, partition_hints: Optional[Dict[str, str]]) -> str:
        if not partition_hints:
            return ""
        formatted_strings = []
        for column_name, template in partition_hints.items():
            formatted_strings.append(
                template.format(column_name=self.sql_client.escape_ddl_identifier(column_name))
            )
        return f"PARTITIONED BY ({', '.join(formatted_strings)})"

    def _iceberg_table_properties(self) -> str:
        table_properties = (
            self.config.table_properties.copy() if self.config.table_properties else {}
        )
        mandatory_properties = {"table_type": "ICEBERG", "format": "parquet"}
        table_properties.update(mandatory_properties)
        return ", ".join([f"'{k}'='{v}'" for k, v in table_properties.items()])

    def manage_lf_tags(self) -> None:
        """Manage Lakeformation tags on the glue database."""
        lf_config = self.config.lakeformation_config

        with boto3_client_lock:
            lf_client = self.config.credentials._to_botocore_session().create_client(
                "lakeformation"
            )

        manager = LfTagsManager(
            lf_client,
            lf_config=lf_config,
            catalog=self.sql_client.catalog_name(quote=False),
            dataset=self.sql_client.dataset_name,
            table=None,
        )
        try:
            manager.process_lf_tags_database()
        except Exception as e:
            logger.error(f"Failed to manage lakeformation tags: {e}")
            raise

    def _get_table_update_sql(
        self, table_name: str, new_columns: Sequence[TColumnSchema], generate_alter: bool
    ) -> List[str]:
        bucket = self.config.staging_config.bucket_url
        dataset = self.sql_client.dataset_name

        sql: List[str] = []

        # for the system tables we need to create empty iceberg tables to be able to run, DELETE and UPDATE queries
        # or if we are in iceberg mode, we create iceberg tables for all tables
        table = self.prepare_load_table(table_name)
        # do not create iceberg tables on staging dataset
        is_iceberg_table = self._is_iceberg_table(table, self.in_staging_dataset_mode)
        columns = ", ".join([self._get_column_def_sql(c, table) for c in new_columns])

        # use qualified table names
        qualified_table_name = self.sql_client.make_qualified_ddl_table_name(table_name)
        if generate_alter:
            # alter table to add new columns at the end
            sql.append(f"""ALTER TABLE {qualified_table_name} ADD COLUMNS ({columns})""")
        else:
            table_prefix = self.table_prefix_layout.format(table_name=table_name)
            if is_iceberg_table:
                partition_clause = self._iceberg_partition_clause(
                    cast(Optional[Dict[str, str]], table.get(PARTITION_HINT))
                )
                stmt = f"""{self._make_create_table(qualified_table_name, table)}
                        ({columns})
                        {partition_clause}"""

                table_location_layout = self.config.table_location_layout
                is_s3_table = self._is_s3_table(self.in_staging_dataset_mode)
                if is_s3_table:
                    logger.info(
                        f"Will create S3 table {table_name} in {self.config.aws_data_catalog}"
                    )
                    if table_location_layout != DEFAULT_TABLE_LOCATION_LAYOUT:
                        logger.info(
                            f"Custom table location layout '{table_location_layout}' is ignored"
                            " because the S3 Tables Catalog manages the table location"
                        )
                else:
                    # create unique tag for iceberg table so it is never recreated in the same folder
                    # athena requires some kind of special cleaning (or that is a bug) so we cannot refresh
                    # iceberg tables without it, by default tag is not added
                    location = f"{bucket}/" + table_location_layout.format(
                        dataset_name=dataset,
                        table_name=table_prefix.rstrip("/"),
                        location_tag=uniq_id(6),
                    )
                    table_properties = self._iceberg_table_properties()
                    logger.info(f"Will create ICEBERG table {table_name} in {location}")
                    stmt += f"""LOCATION '{location.rstrip('/')}'
                        TBLPROPERTIES ({table_properties})"""

                sql.append(stmt)
            else:
                # external tables always here
                location = f"{bucket}/{dataset}/{table_prefix}"
                logger.info(f"Will create EXTERNAL table {table_name} from {location}")
                sql.append(f"""CREATE EXTERNAL TABLE {qualified_table_name}
                        ({columns})
                        STORED AS PARQUET
                        LOCATION '{location}'""")
        return sql

    def update_stored_schema(
        self,
        only_tables: Iterable[str] = None,
        expected_update: TSchemaTables = None,
    ) -> Optional[TSchemaTables]:
        applied_update = super().update_stored_schema(only_tables, expected_update=expected_update)
        # here we could apply tags only if any migration happened, right now we do it on each run
        # NOTE: tags are applied before any data is loaded
        if (
            self.config.lakeformation_config is not None
            and self.config.lakeformation_config.enabled
            is not None  # both True and False are actionable
        ):
            self.manage_lf_tags()
        return applied_update

    def create_load_job(
        self, table: PreparedTableSchema, file_path: str, load_id: str, restore: bool = False
    ) -> LoadJob:
        """Starts SqlLoadJob for files ending with .sql or returns None to let derived classes to handle their specific jobs"""
        job = super().create_load_job(table, file_path, load_id, restore)
        if not job:
            job = (
                FinalizedLoadJobWithFollowupJobs(file_path)
                if self._is_iceberg_table(table)
                else FinalizedLoadJob(file_path)
            )
        return job

    def _create_append_followup_jobs(
        self, table_chain: Sequence[PreparedTableSchema]
    ) -> List[FollowupJobRequest]:
        if self._is_iceberg_table(table_chain[0]):
            return [SqlStagingCopyFollowupJob.from_table_chain(table_chain, self.sql_client)]
        return super()._create_append_followup_jobs(table_chain)

    def _create_replace_followup_jobs(
        self, table_chain: Sequence[PreparedTableSchema]
    ) -> List[FollowupJobRequest]:
        if self._is_iceberg_table(table_chain[0]):
            return [SqlStagingReplaceFollowupJob.from_table_chain(table_chain, self.sql_client)]
        return super()._create_replace_followup_jobs(table_chain)

    def _create_merge_followup_jobs(
        self, table_chain: Sequence[PreparedTableSchema]
    ) -> List[FollowupJobRequest]:
        return [AthenaMergeJob.from_table_chain(table_chain, self.sql_client)]

    def _is_iceberg_table(
        self, table: PreparedTableSchema, is_staging_dataset: bool = False
    ) -> bool:
        table_format = table.get("table_format")
        # all dlt tables that are not loaded via files are iceberg tables, no matter if they are on staging or regular dataset
        # all other iceberg tables are HIVE (external) tables on staging dataset
        table_format_iceberg = table_format == "iceberg"
        return (table_format_iceberg and not is_staging_dataset) or table[
            "write_disposition"
        ] == "skip"

    def _is_s3_table(self, is_staging_dataset: bool = False) -> bool:
        return self.config._is_s3_tables_catalog() and not is_staging_dataset

    def should_load_data_to_staging_dataset(self, table_name: str) -> bool:
        # all iceberg tables need staging
        table = self.prepare_load_table(table_name)
        if self._is_iceberg_table(table):
            return True
        return super().should_load_data_to_staging_dataset(table_name)

    def should_truncate_table_before_load_on_staging_destination(self, table_name: str) -> bool:
        # on athena we only truncate replace tables that are not iceberg
        table = self.prepare_load_table(table_name)
        if table["write_disposition"] == "replace" and not self._is_iceberg_table(table):
            return True
        return False

    def should_drop_table_on_staging_destination(self, dropped_table: TTableSchema) -> bool:
        # in Athena we must drop table in glue and then we must drop data in staging if table is not iceberg
        try:
            existing_table = self.prepare_load_table(dropped_table["name"])
            # do not drop data if new iceberg table got created - storage is handled by Athena
            return not self._is_iceberg_table(existing_table)
        except TableNotFound:
            # table got dropped and is not recreated - drop staging destination
            return True

    def should_load_data_to_staging_dataset_on_staging_destination(self, table_name: str) -> bool:
        """iceberg table data goes into staging on staging destination"""
        table = self.prepare_load_table(table_name)
        if self._is_iceberg_table(table):
            return True
        return super().should_load_data_to_staging_dataset_on_staging_destination(table_name)

    @staticmethod
    def is_dbapi_exception(ex: Exception) -> bool:
        from pyathena.error import Error

        return isinstance(ex, Error)



import re
import sys
from typing import Any, Dict, Type, Union, TYPE_CHECKING, Optional, cast

from dlt.common import logger
from dlt.common.arithmetics import DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE
from dlt.common.data_writers.escape import (
    escape_clickhouse_identifier,
    escape_clickhouse_literal,
    format_clickhouse_datetime_literal,
)
from dlt.common.destination import Destination, DestinationCapabilitiesContext

from dlt.common.destination.typing import PreparedTableSchema
from dlt.common.exceptions import TerminalValueError
from dlt.common.schema.typing import TColumnSchema, TColumnType
from dlt.destinations.type_mapping import TypeMapperImpl
from dlt.destinations.impl.clickhouse.configuration import (
    ClickHouseClientConfiguration,
    ClickHouseCredentials,
)


if TYPE_CHECKING:
    from dlt.destinations.impl.clickhouse.clickhouse import ClickHouseClient
    from clickhouse_driver.dbapi import Connection  # type: ignore[import-untyped]
else:
    Connection = Any


class ClickHouseTypeMapper(TypeMapperImpl):
    sct_to_unbound_dbt = {
        "json": "String",
        "text": "String",
        "double": "Float64",
        "bool": "Boolean",
        "date": "Date",
        "timestamp": "DateTime64(6,'UTC')",
        "time": "String",
        "bigint": "Int64",
        "binary": "String",
        "wei": "Decimal",
    }

    sct_to_dbt = {
        "decimal": "Decimal(%i,%i)",
        "wei": "Decimal(%i,%i)",
        "timestamp": "DateTime64(%i,'UTC')",
    }

    dbt_to_sct = {
        "String": "text",
        "Float64": "double",
        "Bool": "bool",
        "Date": "date",
        "DateTime": "timestamp",
        "DateTime64": "timestamp",
        "Time": "timestamp",
        "Int8": "bigint",
        "Int16": "bigint",
        "Int32": "bigint",
        "Int64": "bigint",
        "Object('json')": "json",
        "Decimal": "decimal",
    }

    def from_destination_type(
        self, db_type: str, precision: Optional[int] = None, scale: Optional[int] = None
    ) -> TColumnType:
        # Remove "Nullable" wrapper.
        db_type = re.sub(r"^Nullable\((?P<type>.+)\)$", r"\g<type>", db_type)

        # Handle DateTime types with and without timezone
        if db_type == "DateTime":
            # Legacy DateTime type without timezone and without precision
            return cast(TColumnType, dict(data_type="timestamp", timezone=False))

        if db_type == "DateTime('UTC')":
            # Legacy DateTime type with UTC timezone
            return cast(TColumnType, dict(data_type="timestamp", timezone=True))

        # Handle DateTime64 types with various formats
        if datetime_match := re.match(
            r"DateTime64(?:\((?P<precision>\d+)(?:,?\s*'(?P<timezone>[^']+)')?\))?",
            db_type,
        ):
            has_timezone = bool(datetime_match["timezone"])

            if datetime_match["precision"]:
                precision = int(datetime_match["precision"])

            column_type = cast(
                TColumnType,
                dict(
                    data_type="timestamp",
                    timezone=has_timezone,
                ),
            )

            if precision is not None:
                column_type["precision"] = precision

            return column_type

        # Extract precision and scale, parameters and remove from string.
        if decimal_match := re.match(
            r"Decimal\((?P<precision>\d+)\s*(?:,\s*(?P<scale>\d+))?\)", db_type
        ):
            precision, scale = decimal_match.groups()
            precision = int(precision)
            scale = int(scale) if scale else 0
            db_type = "Decimal"

        if db_type == "Decimal" and (precision, scale) == self.capabilities.wei_precision:
            return cast(TColumnType, dict(data_type="wei"))

        return super().from_destination_type(db_type, precision, scale)

    def to_db_datetime_type(
        self,
        column: TColumnSchema,
        table: PreparedTableSchema = None,
    ) -> str:
        """Map timestamp type to appropriate ClickHouse datetime type with or without timezone."""
        column_name = column["name"]
        table_name = table.get("name")
        timezone = column.get("timezone", True)
        precision = column.get(
            "precision", self.capabilities.timestamp_precision
        )  # Default precision is 6 for microseconds

        if precision and precision > self.capabilities.max_timestamp_precision:
            logger.warn(
                f"ClickHouse only supports microsecond precision (6) for column '{column_name}' in"
                f" table '{table_name}'. Will use precision 6."
            )
            precision = self.capabilities.max_timestamp_precision

        if timezone:
            return f"DateTime64({precision},'UTC')"
        else:
            return f"DateTime64({precision})"

    def to_db_integer_type(self, column: TColumnSchema, table: PreparedTableSchema = None) -> str:
        """Map integer precision to the appropriate ClickHouse integer type."""
        precision = column.get("precision")
        if precision is None:
            return "Int64"
        if precision <= 8:
            return "Int8"
        if precision <= 16:
            return "Int16"
        if precision <= 32:
            return "Int32"
        if precision <= 64:
            return "Int64"
        raise TerminalValueError(
            f"bigint with `{precision=:}` can't be mapped to ClickHouse integer type"
        )


class clickhouse(Destination[ClickHouseClientConfiguration, "ClickHouseClient"]):
    spec = ClickHouseClientConfiguration

    def _raw_capabilities(self) -> DestinationCapabilitiesContext:
        caps = DestinationCapabilitiesContext()
        caps.preferred_loader_file_format = "jsonl"
        caps.supported_loader_file_formats = ["parquet", "jsonl", "model"]
        caps.preferred_staging_file_format = "jsonl"
        caps.supported_staging_file_formats = ["parquet", "jsonl"]
        caps.type_mapper = ClickHouseTypeMapper

        caps.format_datetime_literal = format_clickhouse_datetime_literal
        caps.escape_identifier = escape_clickhouse_identifier
        caps.escape_literal = escape_clickhouse_literal
        # docs are very unclear https://clickhouse.com/docs/en/sql-reference/syntax
        # taking into account other sources: identifiers are case sensitive
        caps.has_case_sensitive_identifiers = True
        # and store as is in the information schema
        caps.casefold_identifier = str

        # https://stackoverflow.com/questions/68358686/what-is-the-maximum-length-of-a-column-in-clickhouse-can-it-be-modified
        caps.max_identifier_length = 255
        caps.max_column_identifier_length = 255

        # ClickHouse has no max `String` type length.
        caps.max_text_data_type_length = sys.maxsize

        caps.schema_supports_numeric_precision = True
        # Use 'Decimal128' with these defaults.
        # https://clickhouse.com/docs/en/sql-reference/data-types/decimal
        caps.decimal_precision = (DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE)
        # Use 'Decimal256' with these defaults.
        caps.wei_precision = (76, 0)
        caps.timestamp_precision = 6

        # https://clickhouse.com/docs/en/operations/settings/settings#max_query_size
        caps.is_max_query_length_in_bytes = True
        caps.max_query_length = 262144

        # ClickHouse has limited support for transactional semantics, especially for `ReplicatedMergeTree`,
        # the default ClickHouse Cloud engine. It does, however, provide atomicity for individual DDL operations like `ALTER TABLE`.
        # https://clickhouse-driver.readthedocs.io/en/latest/dbapi.html#clickhouse_driver.dbapi.connection.Connection.commit
        # https://clickhouse.com/docs/en/guides/developer/transactional#transactions-commit-and-rollback
        caps.supports_transactions = False
        caps.supports_ddl_transactions = False

        caps.supports_truncate_command = True

        caps.supported_merge_strategies = ["delete-insert", "scd2"]
        caps.supported_replace_strategies = ["truncate-and-insert", "insert-from-staging"]
        caps.enforces_nulls_on_alter = False

        caps.sqlglot_dialect = "clickhouse"

        return caps

    @property
    def client_class(self) -> Type["ClickHouseClient"]:
        from dlt.destinations.impl.clickhouse.clickhouse import ClickHouseClient

        return ClickHouseClient

    def __init__(
        self,
        credentials: Union[ClickHouseCredentials, str, Dict[str, Any], Type[Connection]] = None,
        destination_name: str = None,
        environment: str = None,
        select_sequential_consistency: int = None,
        **kwargs: Any,
    ) -> None:
        """Configure the ClickHouse destination to use in a pipeline.

        All arguments provided here supersede other configuration sources such as environment
        variables and dlt config files.

        Args:
            credentials (Union[ClickHouseCredentials, str, Dict[str, Any], Type[Connection]], optional): Credentials to connect to the clickhouse database. Can be an instance of `ClickHouseCredentials` or
                a connection string in the format `clickhouse://user:password@host:port/database`
            destination_name (str, optional): Name of the destination, can be used in config section to differentiate between multiple of the same type
            environment (str, optional): Environment of the destination
            select_sequential_consistency (int, optional): Ensures read-after-write consistency on ClickHouse Cloud and clusters. Defaults to 1.
            **kwargs (Any): Additional arguments passed to the destination config
        """
        super().__init__(
            credentials=credentials,
            destination_name=destination_name,
            environment=environment,
            select_sequential_consistency=select_sequential_consistency,
            **kwargs,
        )


clickhouse.register()



from typing import (
    Optional,
    ClassVar,
    Iterator,
    Any,
    AnyStr,
    Sequence,
    Tuple,
    Dict,
    Callable,
    Type,
)
from copy import deepcopy
import re

from contextlib import contextmanager
from pendulum.datetime import DateTime, Date
from datetime import datetime  # noqa: I251

import pyathena
from pyathena import connect
from pyathena.connection import Connection
from pyathena.error import OperationalError, DatabaseError, ProgrammingError, IntegrityError, Error
from pyathena.formatter import (
    DefaultParameterFormatter,
    _DEFAULT_FORMATTERS,
    Formatter,
    _format_date,
)

from dlt.common import logger
from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.destination.dataset import DBApiCursor
from dlt.common.data_writers.escape import escape_hive_identifier

from dlt.destinations.typing import DBApi, DBTransaction
from dlt.destinations.exceptions import (
    DatabaseTerminalException,
    DatabaseTransientException,
    DatabaseUndefinedRelation,
)
from dlt.destinations.sql_client import (
    SqlClientBase,
    DBApiCursorImpl,
    raise_database_error,
    raise_open_connection_error,
)
from dlt.destinations.impl.athena.configuration import (
    DEFAULT_AWS_DATA_CATALOG,
    AthenaClientConfiguration,
)


# add a formatter for pendulum to be used by pyathen dbapi
def _format_pendulum_datetime(formatter: Formatter, escaper: Callable[[str], str], val: Any) -> Any:
    # copied from https://github.com/laughingman7743/PyAthena/blob/f4b21a0b0f501f5c3504698e25081f491a541d4e/pyathena/formatter.py#L114
    # https://docs.aws.amazon.com/athena/latest/ug/engine-versions-reference-0003.html#engine-versions-reference-0003-timestamp-changes
    # ICEBERG tables have TIMESTAMP(6), other tables have TIMESTAMP(3), we always generate TIMESTAMP(6)
    # it is up to the user to cut the microsecond part
    val_string = val.strftime("%Y-%m-%d %H:%M:%S.%f")
    return f"""TIMESTAMP '{val_string}'"""


class DLTAthenaFormatter(DefaultParameterFormatter):
    _INSTANCE: ClassVar["DLTAthenaFormatter"] = None

    def __new__(cls: Type["DLTAthenaFormatter"]) -> "DLTAthenaFormatter":
        if cls._INSTANCE:
            return cls._INSTANCE
        return super().__new__(cls)

    def __init__(self) -> None:
        if DLTAthenaFormatter._INSTANCE:
            return
        formatters = deepcopy(_DEFAULT_FORMATTERS)
        formatters[DateTime] = _format_pendulum_datetime
        formatters[datetime] = _format_pendulum_datetime
        formatters[Date] = _format_date

        super(DefaultParameterFormatter, self).__init__(mappings=formatters, default=None)
        DLTAthenaFormatter._INSTANCE = self


class AthenaSQLClient(SqlClientBase[Connection]):
    dbapi: ClassVar[DBApi] = pyathena

    def __init__(
        self,
        dataset_name: str,
        staging_dataset_name: str,
        config: AthenaClientConfiguration,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        super().__init__(None, dataset_name, staging_dataset_name, capabilities)
        self._conn: Connection = None
        self.config = config
        self.credentials = config.credentials

    @raise_open_connection_error
    def open_connection(self) -> Connection:
        # set `use_catalog_name` to False because we use fully qualified table names including
        # catalog in all queries and we may (depending on user config) do cross-catalog queries
        self._conn = connect(
            schema_name=self.dataset_name, **self.config.to_connector_params(use_catalog_name=False)
        )
        return self._conn

    def close_connection(self) -> None:
        self._conn.close()
        self._conn = None

    @property
    def native_connection(self) -> Connection:
        return self._conn

    def catalog_name(
        self, quote: bool = True, casefold: bool = True, ddl: bool = False
    ) -> Optional[str]:
        catalog_name = self.config.aws_data_catalog

        if self.is_staging_dataset_active:
            if self.config.staging_aws_data_catalog is None:
                if self.config._is_s3_tables_catalog():
                    # staging dataset has Parquet tables, which we can't register in S3 Tables
                    # Catalog; hence, use default catalog
                    catalog_name = DEFAULT_AWS_DATA_CATALOG
            else:
                catalog_name = self.config.staging_aws_data_catalog

        if casefold:
            catalog_name = self.capabilities.casefold_identifier(catalog_name)
        if quote:
            escaper = self.escape_ddl_identifier if ddl else self.capabilities.escape_identifier
            catalog_name = escaper(catalog_name)

        return catalog_name

    def _qualify_info_schema_table_name(self, table_name: str) -> str:
        catalog_name = self.catalog_name()
        return f"{catalog_name}.{super()._qualify_info_schema_table_name(table_name)}"

    def escape_ddl_identifier(self, v: str) -> str:
        # https://docs.aws.amazon.com/athena/latest/ug/tables-databases-columns-names.html
        # Athena uses HIVE to create tables but for querying it uses PRESTO (so normal escaping)
        if not v:
            return v
        v = self.capabilities.casefold_identifier(v)
        # athena uses hive escaping
        return escape_hive_identifier(v)

    def fully_qualified_ddl_dataset_name(self) -> str:
        catalog_name = self.catalog_name(ddl=True)
        schema_name = self.escape_ddl_identifier(self.dataset_name)
        return f"{catalog_name}.{schema_name}"

    def make_qualified_ddl_table_name(self, table_name: str) -> str:
        table_name = self.escape_ddl_identifier(table_name)
        return f"{self.fully_qualified_ddl_dataset_name()}.{table_name}"

    def create_dataset(self) -> None:
        db_location_clause = (
            f" LOCATION '{self.config.db_location}'" if self.config.db_location else ""
        )
        # HIVE escaping for DDL
        self.execute_sql(
            f"CREATE DATABASE {self.fully_qualified_ddl_dataset_name()}{db_location_clause}"
        )

    def drop_dataset(self) -> None:
        self.execute_sql(f"DROP DATABASE {self.fully_qualified_ddl_dataset_name()} CASCADE")

    def drop_tables(self, *tables: str) -> None:
        if not tables:
            return
        statements = [
            f"DROP TABLE IF EXISTS {self.make_qualified_ddl_table_name(table)}" for table in tables
        ]
        self.execute_many(statements)

    @contextmanager
    @raise_database_error
    def begin_transaction(self) -> Iterator[DBTransaction]:
        logger.warning(
            "Athena does not support transactions! Each SQL statement is auto-committed separately."
        )
        yield self

    @raise_database_error
    def commit_transaction(self) -> None:
        pass

    @raise_database_error
    def rollback_transaction(self) -> None:
        raise NotImplementedError("You cannot rollback Athena SQL statements.")

    @staticmethod
    def _make_database_exception(ex: Exception) -> Exception:
        if isinstance(ex, OperationalError):
            if "TABLE_NOT_FOUND" in str(ex):
                return DatabaseUndefinedRelation(ex)
            elif "SCHEMA_NOT_FOUND" in str(ex):
                return DatabaseUndefinedRelation(ex)
            elif "Table" in str(ex) and " not found" in str(ex):
                return DatabaseUndefinedRelation(ex)
            elif "Database does not exist" in str(ex):
                return DatabaseUndefinedRelation(ex)
            return DatabaseTerminalException(ex)
        elif isinstance(ex, (ProgrammingError, IntegrityError)):
            return DatabaseTerminalException(ex)
        if isinstance(ex, DatabaseError):
            return DatabaseTransientException(ex)
        return ex

    def execute_sql(
        self, sql: AnyStr, *args: Any, **kwargs: Any
    ) -> Optional[Sequence[Sequence[Any]]]:
        with self.execute_query(sql, *args, **kwargs) as curr:
            if curr.description is None:
                return None
            else:
                f = curr.fetchall()
                return f

    @staticmethod
    def _convert_to_old_pyformat(
        new_style_string: str, args: Tuple[Any, ...]
    ) -> Tuple[str, Dict[str, Any]]:
        # create a list of keys
        keys = ["arg" + str(i) for i, _ in enumerate(args)]
        # create an old style string and replace placeholders
        old_style_string, count = re.subn(
            r"%s", lambda _: "%(" + keys.pop(0) + ")s", new_style_string
        )
        # create a dictionary mapping keys to args
        mapping = dict(zip(["arg" + str(i) for i, _ in enumerate(args)], args))
        # raise if there is a mismatch between args and string
        if count != len(args):
            raise DatabaseTransientException(OperationalError())
        return old_style_string, mapping

    @contextmanager
    @raise_database_error
    def execute_query(self, query: AnyStr, *args: Any, **kwargs: Any) -> Iterator[DBApiCursor]:
        assert isinstance(query, str)
        db_args = kwargs
        # convert sql and params to PyFormat, as athena does not support anything else
        if args:
            query, db_args = self._convert_to_old_pyformat(query, args)
            if kwargs:
                db_args.update(kwargs)

        with self._conn.cursor(formatter=DLTAthenaFormatter()) as cursor:
            for query_line in query.split(";"):
                if query_line.strip():
                    try:
                        cursor.execute(query_line, db_args)
                    # catch key error only here, this will show up if we have a missing parameter
                    except KeyError:
                        raise DatabaseTransientException(OperationalError())

            # TODO: (important) allow to use PandasCursor and ArrowCursor to get fast data access
            #    the problem: you need to set the cursor type upfront. so if user uses wrong cursor
            #    we won't be able to dynamically change it.
            yield DBApiCursorImpl(cursor)



import platform

from dlt.common.destination import DestinationCapabilitiesContext

if platform.python_implementation() == "PyPy":
    import psycopg2cffi as psycopg2
    from psycopg2cffi.sql import SQL, Composed, Composable
else:
    import psycopg2
    from psycopg2.sql import SQL, Composed, Composable

from contextlib import contextmanager
from typing import Any, AnyStr, ClassVar, Iterator, Optional, Sequence

from dlt.destinations.exceptions import (
    DatabaseTerminalException,
    DatabaseTransientException,
    DatabaseUndefinedRelation,
)
from dlt.common.destination.dataset import DBApiCursor

from dlt.destinations.typing import DBApi, DBTransaction
from dlt.destinations.sql_client import (
    DBApiCursorImpl,
    SqlClientBase,
    raise_database_error,
    raise_open_connection_error,
)

from dlt.destinations.impl.postgres.configuration import PostgresCredentials


class Psycopg2SqlClient(SqlClientBase["psycopg2.connection"], DBTransaction):
    dbapi: ClassVar[DBApi] = psycopg2

    def __init__(
        self,
        dataset_name: str,
        staging_dataset_name: str,
        credentials: PostgresCredentials,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        super().__init__(credentials.database, dataset_name, staging_dataset_name, capabilities)
        self._conn: psycopg2.connection = None
        self.credentials = credentials

    def open_connection(self) -> "psycopg2.connection":
        query_options = self.credentials.get_query().get("options", "")
        self._conn = psycopg2.connect(
            dsn=self.credentials.to_native_representation(),
            options=f"{query_options} -csearch_path={self.fully_qualified_dataset_name()},public",
        )
        # we'll provide explicit transactions see _reset
        self._reset_connection()
        return self._conn

    @raise_open_connection_error
    def close_connection(self) -> None:
        if self._conn:
            self._conn.close()
            self._conn = None

    @contextmanager
    def begin_transaction(self) -> Iterator[DBTransaction]:
        try:
            self._conn.autocommit = False
            yield self
            self.commit_transaction()
        except Exception:
            self.rollback_transaction()
            raise

    @raise_database_error
    def commit_transaction(self) -> None:
        self._conn.commit()
        self._conn.autocommit = True

    @raise_database_error
    def rollback_transaction(self) -> None:
        self._conn.rollback()
        self._conn.autocommit = True

    @property
    def native_connection(self) -> "psycopg2.connection":
        return self._conn

    # @raise_database_error
    def execute_sql(
        self, sql: AnyStr, *args: Any, **kwargs: Any
    ) -> Optional[Sequence[Sequence[Any]]]:
        with self.execute_query(sql, *args, **kwargs) as curr:
            if curr.description is None:
                return None
            else:
                f = curr.fetchall()
                return f

    @contextmanager
    @raise_database_error
    def execute_query(self, query: AnyStr, *args: Any, **kwargs: Any) -> Iterator[DBApiCursor]:
        curr: DBApiCursor = None
        db_args = args if args else kwargs if kwargs else None
        with self._conn.cursor() as curr:
            try:
                curr.execute(query, db_args)
                yield DBApiCursorImpl(curr)
            except psycopg2.Error:
                try:
                    self._reset_connection()
                except psycopg2.Error:
                    self.close_connection()
                    self.open_connection()
                raise

    def execute_fragments(
        self, fragments: Sequence[AnyStr], *args: Any, **kwargs: Any
    ) -> Optional[Sequence[Sequence[Any]]]:
        # compose the statements using psycopg2 library
        composed = Composed(sql if isinstance(sql, Composable) else SQL(sql) for sql in fragments)
        return self.execute_sql(composed, *args, **kwargs)

    def _reset_connection(self) -> None:
        # self._conn.autocommit = True
        self._conn.reset()
        self._conn.autocommit = True

    @classmethod
    def _make_database_exception(cls, ex: Exception) -> Exception:
        if isinstance(ex, (psycopg2.errors.UndefinedTable, psycopg2.errors.InvalidSchemaName)):
            return DatabaseUndefinedRelation(ex)
        if isinstance(
            ex,
            (
                psycopg2.OperationalError,
                psycopg2.InternalError,
                psycopg2.errors.SyntaxError,
                psycopg2.errors.UndefinedFunction,
            ),
        ):
            term = cls._maybe_make_terminal_exception_from_data_error(ex)
            if term:
                return term
            else:
                return DatabaseTransientException(ex)
        elif isinstance(
            ex, (psycopg2.DataError, psycopg2.ProgrammingError, psycopg2.IntegrityError)
        ):
            return DatabaseTerminalException(ex)
        elif isinstance(ex, TypeError):
            # psycopg2 raises TypeError on malformed query parameters
            return DatabaseTransientException(psycopg2.ProgrammingError(ex))
        elif cls.is_dbapi_exception(ex):
            return DatabaseTransientException(ex)
        else:
            return ex

    @staticmethod
    def _maybe_make_terminal_exception_from_data_error(
        pg_ex: psycopg2.DataError,
    ) -> Optional[Exception]:
        return None

    @staticmethod
    def is_dbapi_exception(ex: Exception) -> bool:
        return isinstance(ex, psycopg2.Error)



import os
import tomlkit
import tomlkit.exceptions
import tomlkit.items
from typing import Any, Optional, List

from dlt.common.utils import update_dict_nested
from dlt.common.configuration.exceptions import ConfigProviderException

from .doc import BaseDocProvider, CustomLoaderDocProvider

CONFIG_TOML = "config.toml"
SECRETS_TOML = "secrets.toml"


class StringTomlProvider(BaseDocProvider):
    def __init__(self, toml_string: str) -> None:
        super().__init__(StringTomlProvider.loads(toml_string).unwrap())

    # def update(self, toml_string: str) -> None:
    #     self._config_doc = StringTomlProvider.loads(toml_string).unwrap()

    def dumps(self) -> str:
        return tomlkit.dumps(self._config_doc)

    @staticmethod
    def loads(toml_string: str) -> tomlkit.TOMLDocument:
        return tomlkit.parse(toml_string)

    @property
    def supports_secrets(self) -> bool:
        return True

    @property
    def name(self) -> str:
        return "memory"


class SettingsTomlProvider(CustomLoaderDocProvider):
    _config_toml: tomlkit.TOMLDocument
    """Holds tomlkit document with config values that is in sync with _config_doc"""

    def __init__(
        self,
        name: str,
        supports_secrets: bool,
        file_name: str,
        resolvable_dirs: List[str],
    ) -> None:
        """Creates config provider from a `toml` file

        The provider loads the `toml` file with specified name and from specified folder. If `global_dir` is specified,
        it will additionally look for `file_name` in `dlt` global dir (home dir by default) and merge the content.
        The "settings" (`settings_dir`) values overwrite the "global" values.

        If toml file under `settings_dir` is not found it will look into Google Colab userdata object for a value
        with name `file_name` and load toml file from it.
        If that one is not found, it will try to load Streamlit `secrets.toml` file.

        If none of the files exist, an empty provider is created.

        Args:
            name(str): name of the provider when registering in context
            supports_secrets(bool): allows to store secret values in this provider
            file_name (str): The name of `toml` file to load
            resolvable_dirs (List[str]): A list of directories to resolve the file from, files will be merged into each other in the order the directories are specified. Provider is writeable if only one dir specified.

        Raises:
            TomlProviderReadException: File could not be read, most probably `toml` parsing error
        """
        # set supports_secrets early, we need this flag to read config
        self._supports_secrets = supports_secrets
        # read toml file from local or from various environments
        self._toml_paths = self._resolve_toml_paths(
            file_name, [d for d in resolvable_dirs if d is not None]
        )
        # read toml files and set present locations
        self._present_locations: List[str] = []
        self._config_toml = self._read_toml_files(name, file_name, self._toml_paths)

        super().__init__(
            name,
            self._config_toml.unwrap,
            supports_secrets,
            self._toml_paths,
        )

    def _resolve_toml_paths(self, file_name: str, resolvable_dirs: List[str]) -> List[str]:
        return [os.path.join(d, file_name) for d in resolvable_dirs]

    def write_toml(self) -> None:
        assert (
            len(self._toml_paths) == 1
        ), "Will not write configs when more than one toml path was resolved. Found paths: " + str(
            self._toml_paths
        )
        with open(self._toml_paths[0], "w", encoding="utf-8") as f:
            tomlkit.dump(self._config_toml, f)

    def set_value(self, key: str, value: Any, pipeline_name: Optional[str], *sections: str) -> None:
        # write both into tomlkit and dict representations
        ConvertError = (
            tomlkit.items.ConvertError
            if hasattr(tomlkit.items, "ConvertError")
            else tomlkit.items._ConvertError
        )
        try:
            self._set_value(self._config_toml, key, value, pipeline_name, *sections)
        except ConvertError:
            pass
        if hasattr(value, "unwrap"):
            value = value.unwrap()
        super().set_value(key, value, pipeline_name, *sections)

    @property
    def is_empty(self) -> bool:
        return len(self._config_toml.body) == 0 and super().is_empty

    @property
    def present_locations(self) -> List[str]:
        return self._present_locations

    def set_fragment(
        self, key: Optional[str], value_or_fragment: str, pipeline_name: str, *sections: str
    ) -> None:
        # write both into tomlkit and dict representations
        ConvertError = (
            tomlkit.items.ConvertError
            if hasattr(tomlkit.items, "ConvertError")
            else tomlkit.items._ConvertError
        )
        try:
            self._config_toml = self._set_fragment(
                self._config_toml, key, value_or_fragment, pipeline_name, *sections
            )
        except ConvertError:
            pass
        super().set_fragment(key, value_or_fragment, pipeline_name, *sections)

    def to_toml(self) -> str:
        return tomlkit.dumps(self._config_toml)

    def _read_google_colab_secrets(self, name: str, file_name: str) -> tomlkit.TOMLDocument:
        """Try to load the toml from google colab userdata object"""
        try:
            from google.colab import userdata
            from dlt.common.runtime.exec_info import is_notebook

            # make sure we work in interactive mode (get_ipython() is available)
            # when dlt cli is run, userdata is available but without a kernel
            if not is_notebook():
                return None

            try:
                return tomlkit.loads(userdata.get(file_name))
            except (userdata.SecretNotFoundError, userdata.NotebookAccessError):
                # document not found if secret does not exist or we have no permission
                return None
        except ImportError:
            # document not found if google colab context does not exist
            return None

    def _read_streamlit_secrets(self, name: str, file_name: str) -> tomlkit.TOMLDocument:
        """Try to load the toml from Streamlit secrets."""
        # only secrets can come from streamlit
        if not self.supports_secrets:
            return None

        try:
            import streamlit as st
            import streamlit.runtime as st_r

            if not st_r.exists():
                return None

            # Access the entire secrets store
            secrets_ = st.secrets
            if secrets_.load_if_toml_exists():
                # Convert the dictionary to a TOML string
                toml_str = tomlkit.dumps(secrets_.to_dict())

                # Parse the TOML string into a TOMLDocument
                toml_doc = tomlkit.parse(toml_str)
                return toml_doc
            else:
                return None
        except tomlkit.exceptions.TOMLKitError:
            raise
        except Exception:
            # Not in a Streamlit context
            return None

    def _read_toml_file(self, toml_path: str) -> tomlkit.TOMLDocument:
        if os.path.isfile(toml_path):
            with open(toml_path, "r", encoding="utf-8") as f:
                # use whitespace preserving parser
                return tomlkit.load(f)
        else:
            return None

    def _read_toml_files(
        self, name: str, file_name: str, toml_paths: List[str]
    ) -> tomlkit.TOMLDocument:
        """Merge all toml files into one"""

        try:
            # merge all toml files into one
            result_toml: Optional[tomlkit.TOMLDocument] = None
            for path in toml_paths:
                if (loaded_toml := self._read_toml_file(path)) is not None:
                    if result_toml is None:
                        result_toml = loaded_toml
                    else:
                        result_toml = update_dict_nested(loaded_toml, result_toml)
                    # store as present location
                    self._present_locations.append(path)

            # if nothing was found, try to load from google colab or streamlit
            if result_toml is None:
                if (result_toml := self._read_google_colab_secrets(name, file_name)) is not None:
                    pass
                elif (result_toml := self._read_streamlit_secrets(name, file_name)) is not None:
                    pass
                else:
                    result_toml = tomlkit.document()

            return result_toml
        except Exception as ex:
            raise TomlProviderReadException(name, file_name, toml_paths, str(ex))


class ConfigTomlProvider(SettingsTomlProvider):
    def __init__(self, settings_dir: str, global_dir: str = None) -> None:
        super().__init__(CONFIG_TOML, False, CONFIG_TOML, [settings_dir, global_dir])

    @property
    def is_writable(self) -> bool:
        return True


class SecretsTomlProvider(SettingsTomlProvider):
    def __init__(self, settings_dir: str, global_dir: str = None) -> None:
        super().__init__(SECRETS_TOML, True, SECRETS_TOML, [settings_dir, global_dir])

    @property
    def is_writable(self) -> bool:
        return True


class TomlProviderReadException(ConfigProviderException):
    def __init__(
        self, provider_name: str, file_name: str, full_paths: List[str], toml_exception: str
    ) -> None:
        self.file_name = file_name
        self.full_paths = full_paths
        msg = f"A problem encountered when loading {provider_name} from paths {full_paths}:\n"
        msg += toml_exception
        super().__init__(provider_name, msg)



import os
import dataclasses
import sys
from urllib.parse import urlencode
from typing import Any, ClassVar, Dict, Final, List, TYPE_CHECKING

from dlt.common.configuration.specs.connection_string_credentials import ConnectionStringCredentials
from dlt.version import __version__
from dlt.common.configuration import configspec
from dlt.common.destination.client import DestinationClientDwhWithStagingConfiguration
from dlt.common.destination.exceptions import DestinationTerminalException
from dlt.common.typing import TSecretStrValue
from dlt.common.utils import digest128

from dlt.destinations.impl.duckdb.configuration import DuckDbBaseCredentials, DuckDbConnectionPool

if TYPE_CHECKING:
    from duckdb import DuckDBPyConnection
else:
    DuckDBPyConnection = Any  # type: ignore[assignment,misc]

MOTHERDUCK_DRIVERNAME = "md"
MOTHERDUCK_USER_AGENT = f"dlt/{__version__}({sys.platform})"
MOTHERDUCK_DEFAULT_TOKEN_ENV = "motherduck_token"


class MotherduckConnectionPool(DuckDbConnectionPool):
    def borrow_conn(
        self,
        global_config: Dict[str, Any] = None,
        local_config: Dict[str, Any] = None,
        pragmas: List[str] = None,
    ) -> DuckDBPyConnection:
        from duckdb import HTTPException, InvalidInputException

        try:
            return super().borrow_conn(global_config, local_config, pragmas)
        except (InvalidInputException, HTTPException) as ext_ex:
            if "Failed to download extension" in str(ext_ex) and "motherduck" in str(ext_ex):
                from importlib.metadata import version as pkg_version

                raise MotherduckLocalVersionNotSupported(pkg_version("duckdb")) from ext_ex

            raise


@configspec(init=False)
class MotherDuckCredentials(DuckDbBaseCredentials, ConnectionStringCredentials):
    drivername: Final[str] = dataclasses.field(  # type: ignore
        default="md", init=False, repr=False, compare=False
    )
    username: str = "motherduck"
    password: TSecretStrValue = None
    database: str = "my_db"
    custom_user_agent: str = MOTHERDUCK_USER_AGENT

    __config_gen_annotations__: ClassVar[List[str]] = ["password", "database"]

    def _conn_str(self) -> str:
        # TODO: fix dbt profile.yml to receive full conn str
        _str = f"{MOTHERDUCK_DRIVERNAME}:{self.database}"

        q_ = dict(self.query or {})
        if self.password:
            q_["motherduck_token"] = self.password

        return _str + "?" + urlencode(q_)

    def _token_to_password(self) -> None:
        if self.query:
            # backward compat
            if "token" in self.query:
                self.password = self.query.pop("token")
            if "motherduck_token" in self.query:
                self.password = self.query.pop("motherduck_token")

    def parse_native_representation(self, native_value: Any) -> None:
        if isinstance(native_value, str):
            # https://motherduck.com/docs/key-tasks/authenticating-and-connecting-to-motherduck/authenticating-to-motherduck/#storing-the-access-token-as-an-environment-variable
            # ie. md:dlt_data_3?motherduck_token=<my service token>
            if native_value.startswith("md:") and not native_value.startswith("md:/"):
                native_value = "md:///" + native_value[3:]  # skip md:
        super().parse_native_representation(native_value)
        self._token_to_password()

    def on_partial(self) -> None:
        """Takes a token from query string and reuses it as a password"""
        self._token_to_password()
        if not self.is_partial() or self._has_default_token():
            self.resolve()

    def on_resolved(self) -> None:
        """Adds custom agent to global config"""
        if self.global_config is None:
            self.global_config = {}
        self.global_config["custom_user_agent"] = self.custom_user_agent or MOTHERDUCK_USER_AGENT
        self.conn_pool = MotherduckConnectionPool(self)

    def _has_default_token(self) -> bool:
        # TODO: implement default connection interface
        return (
            MOTHERDUCK_DEFAULT_TOKEN_ENV in os.environ
            or MOTHERDUCK_DEFAULT_TOKEN_ENV.upper() in os.environ
        )


@configspec
class MotherDuckClientConfiguration(DestinationClientDwhWithStagingConfiguration):
    destination_type: Final[str] = dataclasses.field(  # type: ignore
        default="motherduck", init=False, repr=False, compare=False
    )
    credentials: MotherDuckCredentials = None

    create_indexes: bool = (
        False  # should unique indexes be created, this slows loading down massively
    )

    def fingerprint(self) -> str:
        """Returns a fingerprint of user access token"""
        if self.credentials and self.credentials.password:
            return digest128(self.credentials.password)
        return ""


class MotherduckLocalVersionNotSupported(DestinationTerminalException):
    def __init__(self, duckdb_version: str) -> None:
        self.duckdb_version = duckdb_version
        super().__init__(
            f"Looks like your local duckdb version ({duckdb_version}) is not supported by"
            " Motherduck"
        )
